#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module provides support for KeyStore used in MasterBootImage."""

from typing import Optional

from Crypto.Cipher import AES

from spsdk import SPSDKError
from spsdk.utils.easy_enum import Enum


class KeySourceType(Enum):
    """Device key source."""

    OTP = (0, "OTP", "Device keys stored in OTP")
    KEYSTORE = (1, "KEYSTORE", "Device keys stored in KeyStore")


class KeyStore:
    """Provide info about KeyStore for MaterBootImage."""

    # size of key store in bytes
    KEY_STORE_SIZE = (
        1424  # Size can be device-specific, the current value is valid for RT5xx and RT6xx
    )

    SBKEK_SIZE = 32  # Size of Secure Binary KEK in bytes

    OTP_MASTER_KEY_SIZE = 32  # Size of OTP master key in bytes
    OTFAD_KEY_SIZE = 16  # Size of OTFAD key in bytes

    @property
    def key_source(self) -> KeySourceType:
        """Device key source."""
        return self._key_source

    def __init__(self, key_source: KeySourceType, key_store: Optional[bytes] = None) -> None:
        """Initialize Keystore.

        :param key_source: device key source
        :param key_store: initial content of the key store in the bootable image; None if empty
        :raises SPSDKError: If invalid key-store size
        :raises SPSDKError: KeyStore can be initialized only if key_source == KEYSTORE
        """
        if key_store:
            if len(key_store) != self.KEY_STORE_SIZE:
                raise SPSDKError(
                    f"Invalid key-store size, expected is {str(self.KEY_STORE_SIZE)} bytes"
                )
            if key_source != KeySourceType.KEYSTORE:
                raise SPSDKError("KeyStore can be initialized only if key_source == KEYSTORE")

        self._key_source = key_source
        self._key_store = key_store

    def export(self) -> bytes:
        """Binary key store content; empty bytes for empty key-store."""
        return self._key_store if self._key_store else bytes()

    def info(self) -> str:
        """Information about key store in text form."""
        return (
            f"Device key source:    {KeySourceType.name(self.key_source)}\n"
            f"Device key store len: {str(len(self.export()))}"
        )

    @staticmethod
    def derive_hmac_key(hmac_key: bytes) -> bytes:
        """Derive HMAC from master or user key.

        :param hmac_key: either master-key (for key_source == OTP) or user key (for key_source == KEYSTORE)
        :return: key used for image header authentication in LoadToRam images
        :raises SPSDKError: If invalid length of hmac key
        """
        if len(hmac_key) != 32:
            raise SPSDKError("Invalid length of hmac key")
        aes = AES.new(hmac_key, AES.MODE_ECB)
        return aes.encrypt(bytes([0] * 16))

    @staticmethod
    def derive_enc_image_key(master_key: bytes) -> bytes:
        """Derive "enc_image_key" from master key.

        :param master_key: stored in OTP
        :return: key used to decrypt encrypted images during boot
        :raises SPSDKError: If invalid length of master key
        """
        if len(master_key) != 32:
            raise SPSDKError("Invalid length of master key")
        aes = AES.new(master_key, AES.MODE_ECB)
        return aes.encrypt(bytes([1] + [0] * 15 + [2] + [0] * 15))

    @staticmethod
    def derive_sb_kek_key(master_key: bytes) -> bytes:
        """Derive SBKEK key from master key.

        :param master_key: 32 bytes key, stored in OTP
        :return: encryption key to handle SB2 file (update capsule)
        :raises SPSDKError: If invalid length of master key
        """
        if len(master_key) != 32:
            raise SPSDKError("Invalid length of master key")
        aes = AES.new(master_key, AES.MODE_ECB)
        return aes.encrypt(bytes([3] + [0] * 15 + [4] + [0] * 15))

    @staticmethod
    def derive_otfad_kek_key(master_key: bytes, otfad_input: bytes) -> bytes:
        """Derive OTFAD KEK key from master key and OTFAD input.

        :param master_key: 32 bytes key, stored in OTP
        :param otfad_input: 16 bytes input, stored in OTP
        :return: OTFAD encryption key for FLASH encryption/decryption
        :raises SPSDKError: If invalid length of master key
        :raises SPSDKError: If invalid length of input
        """
        if len(master_key) != KeyStore.OTP_MASTER_KEY_SIZE:
            raise SPSDKError("Invalid length of master key")
        if len(otfad_input) != KeyStore.OTFAD_KEY_SIZE:
            raise SPSDKError("Invalid length of input")
        aes = AES.new(master_key, AES.MODE_ECB)
        return aes.encrypt(otfad_input)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing command (BD) file parser."""

import logging
from numbers import Number
from typing import Any, Dict, List, Optional

from sly import Parser
from sly.lex import Token
from sly.yacc import YaccProduction

from spsdk import SPSDKError

from . import sly_bd_lexer as bd_lexer


# pylint: disable=too-many-public-methods,too-many-lines
# too-many-public-methods : every method in the parser represents a syntax rule,
#   this is necessary and thus can't be omitted. From this perspective this check
#   is disabled.
# too-many-lines : the class can't be shortened, as all the methods represent
#   rules.
class BDParser(Parser):
    """Command (BD) file parser for elftosb utility.

    The parser is based on SLY framework (python implementation of Lex/YACC)
    and is used to parse the command file, which serves as an input for elftosb
    utility to create a secure binary in 2.1 format.
    The parser doesn't support the grammar of legacy elftosb fully at the moment.
    See the documentation for details.
    """

    # Import tokens from lexer. This is required by the parser!
    tokens = bd_lexer.BDLexer.tokens
    # tokens = BDLexer.tokens

    # Uncomment this line to output parser debug file
    # debugfile = "parser.out"

    log = logging.getLogger(__name__)
    log.setLevel(logging.ERROR)

    def __init__(self) -> None:
        """Initialization method."""
        super().__init__()
        self._variables: List[bd_lexer.Variable] = []
        self._sources: List[bd_lexer.Variable] = []
        self._keyblobs: List[Dict] = []
        self._sections: List[bd_lexer.Variable] = []
        self._input: Any = None
        self._bd_file: Dict = {}
        self._parse_error: bool = False
        self._extern: List[str] = []
        self._lexer = bd_lexer.BDLexer()

    def _cleanup(self) -> None:
        """Cleans up allocated resources before next parsing."""
        self._variables = []
        self._keyblobs = []
        self._sections = []
        # for some strange reason, mypy assumes this is a redefinition of _input
        self._input = None  # type: ignore
        self._bd_file = {}
        self._parse_error = False
        self._lexer.cleanup()

    def parse(
        self, text: str, extern: Optional[List] = None
    ) -> Optional[Dict]:  # pylint: disable=arguments-differ
        """Parse the `input_text` and returns a dictionary of the file content.

        :param text: command file to be parsed in string format
        :param extern: additional files defined on command line during elftosb invocation

        :return: dictionary of the command file content or None on Syntax error
        """
        self._cleanup()
        self._extern = extern or []
        # for some strange reason, mypy assumes this is a redefinition of _input
        self._input: Any = text  # type: ignore

        super().parse(self._lexer.tokenize(text))

        if self._parse_error is True:
            print("BD file parsing not successful.")
            return None

        return self._bd_file

    # Operators precedence
    precedence = (
        ("left", "LOR"),
        ("left", "LAND"),
        ("left", "OR"),
        ("left", "XOR"),
        ("left", "AND"),
        ("left", "EQ", "NE"),
        ("left", "GT", "GE", "LT", "LE"),
        ("left", "LSHIFT", "RSHIFT"),
        ("left", "PLUS", "MINUS"),
        ("left", "TIMES", "DIVIDE", "MOD"),
        ("right", "SIZEOF"),
        ("right", "LNOT", "NOT"),
    )

    # Grammar related stuff
    # TODO remove or is necessary to keep??
    allowed_option_identifiers = (
        "alignment",
        "buildNumber",
        "cleartext",
        "componentVersion",
        "driveTag",
        "flags",
        "productVersion",
        "secinfoClear",
        "sectionFlags",
        "secureBinaryVersion",
        "toolset",
    )
    allowed_sources_option_identifiers = (
        "secinfoClear",
        "toolset",
    )

    # Although the grammar defines a set of reserved keywords, not all of them
    # are checked during parsing, so we have to maintain  two lists. One for the
    # parser and one for future check by elftosb, if we want the grammar to be
    # the same
    # TODO remove or is necessary to keep??
    non_grammar_reserved = (
        "start",
        "end",
        "key",
        "counter",
        "noByteSwap",
    )

    # pylint: disable=undefined-variable,function-redefined,no-self-use,unused-argument
    # undefined-variable : the module uses underscore decorator to define
    #   each rule, however, this causes issues to mypy and pylint.
    # function-redefined : each rule is identified by a function name and a
    #   decorator. However from code checking tools perspective, this is
    #   function redefinition. Thus we need to disable this rule as well.
    # no-self-use : all 'rules' must be class methods, although they don't use
    #   self. Thus we need to omit this rule.
    # unused-argument : not all token input arguments are always used, especially
    #   in rules which are not supported.
    @_("pre_section_block section_block")  # type: ignore
    def command_file(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        """
        token.pre_section_block.update(token.section_block)
        self._bd_file.update(token.pre_section_block)

    @_("pre_section_block options_block")  # type: ignore
    def pre_section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary defining the presection_block.
        """
        options = token.pre_section_block.get("options", {})
        options.update(token.options_block["options"])
        token.pre_section_block["options"] = options
        return token.pre_section_block

    @_("pre_section_block constants_block", "pre_section_block sources_block")  # type: ignore
    def pre_section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary defining the presection block.
        """
        token.pre_section_block.update(token[1])
        return token.pre_section_block

    @_("pre_section_block keyblob_block")  # type: ignore
    def pre_section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary defining the presection block.
        """
        if token.pre_section_block.get("keyblobs") is None:
            token.pre_section_block["keyblobs"] = []
        token.pre_section_block["keyblobs"].append(token.keyblob_block)
        return token.pre_section_block

    @_("empty")  # type: ignore
    def pre_section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary defining the presection block.
        """
        return token.empty

    @_("OPTIONS LBRACE option_def RBRACE")  # type: ignore
    def options_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary defining the options block.
        """
        return token.option_def

    @_("option_def IDENT ASSIGN const_expr SEMI")  # type: ignore
    def option_def(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding an option definition.
        """
        # it appears, that in the option block anything can be defined, so
        # we don't check, whether the identifiers defined there are from the
        # allowed options anymore. The code is left just as a reminder.
        # identifier = token.IDENT
        # if identifier in self.allowed_option_identifiers:
        #     self._variables.append(self.Variable(token.IDENT, "option", token.const_expr))
        #     token.option_def["options"].update({token.IDENT : token.const_expr})
        #     return token.option_def
        # else:
        #     column = BDParser._find_column(self._input, token)
        #     print(f"Unknown option in options block at {token.lineno}/{column}: {token.IDENT}")
        #     self.error(token)
        self._variables.append(bd_lexer.Variable(token.IDENT, "option", token.const_expr))
        token.option_def["options"].update({token.IDENT: token.const_expr})
        return token.option_def

    @_("empty")  # type: ignore
    def option_def(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding an empty option definition.
        """
        return {"options": {}}

    @_("CONSTANTS LBRACE constant_def RBRACE")  # type: ignore
    def constants_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        For now, we don't store the constants in the final bd file.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of constants block.
        """
        dictionary: Dict = {}
        return dictionary

    @_("constant_def IDENT ASSIGN bool_expr SEMI")  # type: ignore
    def constant_def(self, token: YaccProduction):  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        """
        self._variables.append(bd_lexer.Variable(token.IDENT, "constant", token.bool_expr))

    @_("empty")  # type: ignore
    def constant_def(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding empty constant definition.
        """
        return token.empty

    @_("SOURCES LBRACE source_def RBRACE")  # type: ignore
    def sources_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't store the sources in the final BD file for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the definition of sources
        """
        sources = {}
        for source in self._lexer._sources:
            sources[source.name] = source.value
        return {"sources": sources}

    @_("source_def IDENT ASSIGN source_value SEMI")  # type: ignore
    def source_def(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        """
        new_source = bd_lexer.Variable(token.IDENT, "source", token.source_value)
        self._lexer.add_source(new_source)

    @_("source_def IDENT ASSIGN source_value LPAREN source_attr_list RPAREN SEMI")  # type: ignore
    def source_def(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        """
        # TODO add source_attr_list in next stage of development
        # self._sources.append(self.Variable(token.IDENT, "source", token.source_value))
        error_token = Token()
        error_token.lineno = token.lineno
        error_token.index = token._slice[4].index
        self.error(error_token, ": attribute list is not supported")

    @_("empty")  # type: ignore
    def source_def(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding empty content.
        """
        return token.empty

    @_("STRING_LITERAL")  # type: ignore
    def source_value(self, token: YaccProduction) -> str:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: string defining the source value.
        """
        # Everything we read is a string. But strings already contain double quotes,
        # from this perspective we need to remove them, this omit the first and last
        # character.
        return token.STRING_LITERAL[1:-1]

    @_("EXTERN LPAREN int_const_expr RPAREN")  # type: ignore
    def source_value(self, token: YaccProduction) -> str:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: string defining a path defined on command line.
        """
        if token.int_const_expr > len(self._extern) - 1:
            self.error(token, ": extern() out of range")
            return ""
        return self._extern[token.int_const_expr]

    @_("source_attr COMMA source_attr_list")  # type: ignore
    def source_attr_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: empty dictionary as this is not supported right now.
        """
        # TODO support syntax maybe in future if needed???
        dictionary = {}
        return dictionary

    @_("source_attr")  # type: ignore
    def source_attr_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: returns dictionary holding content of source attribute.
        """
        return token.source_attr

    @_("empty")  # type: ignore
    def source_attr_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: returns dictionary holding content of empty source attribute list.
        """
        return {}

    @_("IDENT ASSIGN const_expr")  # type: ignore
    def source_attr(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of source file attributes.
        """
        return {token.IDENT: token.const_expr}

    @_("KEYBLOB LPAREN int_const_expr RPAREN LBRACE keyblob_contents RBRACE")  # type: ignore
    def keyblob_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of keyblob block.
        """
        dictionary = {"keyblob_id": token.int_const_expr, "keyblob_content": token.keyblob_contents}
        dictionary["keyblob_id"] = token.int_const_expr
        dictionary["keyblob_content"] = token.keyblob_contents
        self._keyblobs.append(dictionary)
        return dictionary

    # The legacy tool allowed to have multiple definitions inside a keyblob.
    # It has been agreed, that this makes no sense and may be dangerous.
    # However, it may happen, that someone comes with a use cases, where legacy
    # grammar is needed, thus the code has been left untouched just in case.
    # @_("keyblob_contents LPAREN keyblob_options_list RPAREN")
    # def keyblob_contents(self, token):
    #     l = token.keyblob_contents

    #     # Append only non-empty options lists to simplify further processing
    #     if len(token.keyblob_options_list) != 0:
    #         l.append(token.keyblob_options_list)
    #     return l

    # @_("empty")
    # def keyblob_contents(self, token):
    #     return []

    # @_("keyblob_options")
    # def keyblob_options_list(self, token):
    #     return token.keyblob_options

    # @_("empty")
    # def keyblob_options_list(self, token):
    #     # After discussion internal discussion, we will ignore empty definitions in keyblob
    #     # It's not clear, whether this has some effect on the final sb file or not.
    #     # C++ elftosb implementation is able to parse the file even without empty
    #     # parenthesis
    #     return token.empty

    # @_("IDENT ASSIGN const_expr COMMA keyblob_options")
    # def keyblob_options(self, token):
    #     d = {}
    #     d[token.IDENT] = token.const_expr
    #     d.update(token.keyblob_options)
    #     return d

    # @_("IDENT ASSIGN const_expr")
    # def keyblob_options(self, token):
    #     d = {}
    #     d[token.IDENT] = token.const_expr
    #     return d

    # New keyblob grammar!
    @_("LPAREN keyblob_options RPAREN")  # type: ignore
    def keyblob_contents(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list containing options of each keyblob.
        """
        list_ = [token.keyblob_options]

        return list_

    @_("IDENT ASSIGN const_expr COMMA keyblob_options")  # type: ignore
    def keyblob_options(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of keyblob options.
        """
        dictionary = {}
        dictionary[token.IDENT] = token.const_expr
        dictionary.update(token.keyblob_options)
        return dictionary

    @_("IDENT ASSIGN const_expr")  # type: ignore
    def keyblob_options(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the keyblob option.
        """
        dictionary = {}
        dictionary[token.IDENT] = token.const_expr
        return dictionary

    @_("section_block SECTION LPAREN int_const_expr section_options RPAREN section_contents")  # type: ignore
    def section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a section.
        """
        self._sections.append(
            {
                "section_id": token.int_const_expr,
                "options": token.section_options,
                "commands": token.section_contents,
            }
        )
        token.section_block["sections"] += [
            {
                "section_id": token.int_const_expr,
                "options": token.section_options,
                "commands": token.section_contents,
            }
        ]
        return token.section_block

    @_("empty")  # type: ignore
    def section_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding content of empty section.
        """
        token.empty["sections"] = []
        return token.empty

    @_("SEMI section_option_list")  # type: ignore
    def section_options(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of section options.
        """
        return token.section_option_list

    @_("SEMI")  # type: ignore
    def section_options(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of empty section options.
        """
        dictionary = {}
        return dictionary

    @_("empty")  # type: ignore
    def section_options(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of empty section options.
        """
        return token.empty

    @_("section_option_list COMMA section_option")  # type: ignore
    def section_option_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of section options.
        """
        options = {}
        options.update(token.section_option)
        if token.section_option_list:
            token.section_option_list.append(options)
        return token.section_option_list

    @_("section_option")  # type: ignore
    def section_option_list(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding a section option.
        """
        return [token.section_option]

    @_("IDENT ASSIGN const_expr")  # type: ignore
    def section_option(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a section option.
        """
        return {token.IDENT: token.const_expr}

    @_("LBRACE statement RBRACE")  # type: ignore
    def section_contents(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the section statements.
        """
        return token.statement

    @_("LE SOURCE_NAME SEMI")  # type: ignore
    def section_contents(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        """
        # TODO investigate what does <= SOURCE_NAME does exactly
        self.error(token, ": <= <source_name> syntax is not supported right now.")

    @_("statement basic_stmt SEMI")  # type: ignore
    def statement(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list holding section statements.
        """
        list_ = [] + token.statement
        list_.append(token.basic_stmt)
        return list_

    @_("statement from_stmt")  # type: ignore
    def statement(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support from_stmt for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of from_stmt.
        """
        dictionary = {}
        return dictionary

    @_("statement if_stmt")  # type: ignore
    def statement(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support if statements for now.

        :param token: object holding the content defined in decorator.
        """
        # return token.statement + token.if_stmt

    @_("statement encrypt_block")  # type: ignore
    def statement(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list containing the encrypt statement.
        """
        list_ = [] + token.statement
        list_.append(token.encrypt_block)
        return list_

    @_("statement keywrap_block")  # type: ignore
    def statement(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list containing the keywrap statement.
        """
        list_ = [] + token.statement
        list_.append(token.keywrap_block)
        return list_

    @_("empty")  # type: ignore
    def statement(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: empty list.
        """
        # return empty statement list
        return []

    @_("KEYWRAP LPAREN int_const_expr RPAREN LBRACE LOAD BINARY_BLOB GT int_const_expr SEMI RBRACE")  # type: ignore
    def keywrap_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the keywrap block content.
        """
        dictionary = {"keywrap": {"keyblob_id": token.int_const_expr0}}
        load_cmd = {"load": {"address": token.int_const_expr1, "values": token.BINARY_BLOB}}
        dictionary["keywrap"].update(load_cmd)
        return dictionary

    @_("ENCRYPT LPAREN int_const_expr RPAREN LBRACE load_stmt SEMI RBRACE")  # type: ignore
    def encrypt_block(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the encrypt block content.
        """
        dictionary = {"encrypt": {"keyblob_id": token.int_const_expr}}
        dictionary["encrypt"].update(token.load_stmt)
        return dictionary

    @_(  # type: ignore
        "load_stmt",
        "call_stmt",
        "jump_sp_stmt",
        "mode_stmt",
        "message_stmt",
        "erase_stmt",
        "enable_stmt",
        "reset_stmt",
        "keystore_stmt",
        "version_stmt",
    )
    def basic_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of defined statements.
        """
        return token[0]

    @_("LOAD load_opt load_data load_target")  # type: ignore
    def load_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a load statement.
        """
        # pattern with load options means load -> program command
        if token.load_data.get("pattern") is not None and token.load_opt.get("load_opt") is None:
            cmd = "fill"
        else:
            cmd = "load"
        dictionary: Dict = {cmd: {}}
        dictionary[cmd].update(token.load_opt)
        dictionary[cmd].update(token.load_data)
        dictionary[cmd].update(token.load_target)
        return dictionary

    @_("empty")  # type: ignore
    def load_opt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load options.
        """
        return token.empty

    @_("'@' int_const_expr")  # type: ignore
    def load_opt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load options.
        """
        return {"load_opt": token.int_const_expr}

    @_("IDENT")  # type: ignore
    def load_opt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load options.
        """
        return {"load_opt": token.IDENT}

    @_("int_const_expr")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        if isinstance(token.int_const_expr, str):
            self.error(token, f": identifier '{token.int_const_expr}' is not a source identifier.")
            retval = {"N/A": "N/A"}
        else:
            retval = {"pattern": token.int_const_expr}

        return retval

    @_("STRING_LITERAL")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        return {"file": token.STRING_LITERAL[1:-1]}

    @_("SOURCE_NAME")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        for source in self._lexer._sources:
            if token.SOURCE_NAME == source.name:
                return {"file": source.value}

        # with current implementation, this code won't be ever reached. In case
        # a not defined source file is used as `load_data`, the parser detects
        # it as a different rule:
        #
        # load_data ::= int_const_expr
        #
        # which evaluates as false... however, this fragment is left just in
        # in case something changes.
        self.error(token, ": source file not defined")
        return {"file": "N/A"}

    @_("section_list")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        self.error(token, ": section list is not supported")
        dictionary = {}
        return dictionary

    @_("section_list FROM SOURCE_NAME")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        self.error(token, "section list using from is not supported")
        dictionary = {}
        return dictionary

    @_("BINARY_BLOB")  # type: ignore
    def load_data(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load data.
        """
        # no_spaces = "".join(token.BINARY_BLOB.split())

        return {"values": token.BINARY_BLOB}

    @_("GT PERIOD")  # type: ignore
    def load_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the load target.
        """
        self.error(token, ": '.' as load destination is not supported right now")
        dictionary = {}
        return dictionary

    @_("GT address_or_range")  # type: ignore
    def load_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of load target.
        """
        return token.address_or_range

    @_("empty")  # type: ignore
    def load_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: empty dictionary.
        """
        self.error(token, ": empty load target is not supported right now.")
        return token.empty

    @_("ERASE mem_opt address_or_range")  # type: ignore
    def erase_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of erase statement.
        """
        dictionary: Dict = {token.ERASE: {}}
        dictionary[token.ERASE].update(token.address_or_range)
        dictionary[token.ERASE].update(token.mem_opt)
        return dictionary

    @_("ERASE mem_opt ALL")  # type: ignore
    def erase_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of erase statement.
        """
        dictionary: Dict = {token.ERASE: {"address": 0x00, "flags": 0x01}}
        dictionary[token.ERASE].update(token.mem_opt)
        return dictionary

    @_("ERASE UNSECURE ALL")  # type: ignore
    def erase_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of erase statement.
        """
        return {"erase": {"address": 0x00, "flags": 0x02}}

    @_("ENABLE mem_opt int_const_expr")  # type: ignore
    def enable_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of enable statement.
        """
        dictionary: Dict = {token.ENABLE: {}}
        dictionary[token.ENABLE].update(token.mem_opt)
        dictionary[token.ENABLE]["address"] = token.int_const_expr
        return dictionary

    @_("section_list COMMA section_ref")  # type: ignore
    def section_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the section list content.
        """
        dictionary = {}
        return dictionary

    @_("section_ref")  # type: ignore
    def section_list(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a section reference.
        """
        return token.section_ref

    @_("NOT SECTION_NAME")  # type: ignore
    def section_ref(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a section reference.
        """
        self.error(token, ": section reference is not supported.")
        dictionary = {}
        return dictionary

    @_("SECTION_NAME")  # type: ignore
    def section_ref(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a section reference.
        """
        self.error(token, ": section reference is not supported.")
        return {token.SECTION_NAME}

    @_("int_const_expr")  # type: ignore
    def address_or_range(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of an address.
        """
        address_start = token.int_const_expr
        return {"address": address_start, "length": 1}

    @_("int_const_expr RANGE int_const_expr")  # type: ignore
    def address_or_range(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of an address range.
        """
        address_start = token.int_const_expr0
        length = token.int_const_expr1 - address_start
        return {"address": address_start, "length": length}

    @_("SOURCE_NAME QUESTIONMARK COLON IDENT")  # type: ignore
    def symbol_ref(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        """
        self.error(token, ": symbol reference is not supported.")

    @_("call_type call_target call_arg")  # type: ignore
    def call_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a call statement.
        """
        dictionary: Dict = {token.call_type: {}}
        dictionary[token.call_type].update(token.call_target)
        dictionary[token.call_type].update(token.call_arg)
        return dictionary

    @_("CALL", "JUMP")  # type: ignore
    def call_type(self, token: YaccProduction) -> str:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: string representing 'call' or 'jump'
        """
        return token[0]

    @_("int_const_expr")  # type: ignore
    def call_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a call_target.
        """
        return {"address": token.int_const_expr}

    @_("SOURCE_NAME")  # type: ignore
    def call_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a call target.
        """
        self.error(token, ": source name as call target is not supported.")
        dictionary = {}
        return dictionary

    @_("symbol_ref")  # type: ignore
    def call_target(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a call target.
        """
        self.error(token, ": symbol reference as call target is not supported.")
        dictionary = {}
        return dictionary

    @_("LPAREN RPAREN")  # type: ignore
    def call_arg(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding an empty call argument.
        """
        dictionary = {}
        return dictionary

    @_("LPAREN int_const_expr RPAREN")  # type: ignore
    def call_arg(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding a call argument.
        """
        return {"argument": token.int_const_expr}

    @_("empty")  # type: ignore
    def call_arg(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding an empty call argument.
        """
        return token.empty

    @_("JUMP_SP int_const_expr call_target call_arg")  # type: ignore
    def jump_sp_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content jump statement.
        """
        dictionary: Dict = {"jump": {}}
        dictionary["jump"]["spreg"] = token.int_const_expr
        dictionary["jump"].update(token.call_target)
        dictionary["jump"].update(token.call_arg)
        return dictionary

    @_("RESET")  # type: ignore
    def reset_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of a reset statement.
        """
        return {"reset": {}}

    @_("FROM SOURCE_NAME LBRACE in_from_stmt RBRACE")  # type: ignore
    def from_stmt(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        """
        self.error(token, ": from statement not supported.")

    @_("basic_stmt SEMI")  # type: ignore
    def in_from_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list of statements.
        """
        return token.basic_stmt

    @_("if_stmt")  # type: ignore
    def in_from_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list of statements.
        """
        return token.if_stmt

    @_("empty")  # type: ignore
    def in_from_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: empty list.
        """
        return []

    @_("MODE int_const_expr")  # type: ignore
    def mode_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return:
        """
        self.error(token, ": mode statement is not supported")
        dictionary: Dict = {}
        return dictionary

    @_("message_type STRING_LITERAL")  # type: ignore
    def message_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the message statement.
        """
        # TODO how to handle this statement?? Just print the output??
        # We return empty dict just not to get error while the parser
        # tries to add this statement into a list...
        dictionary: Dict = {}
        return dictionary

    @_("INFO", "WARNING", "ERROR")  # type: ignore
    def message_type(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: empty dictionary.
        """
        self.error(token, ": info/warning/error messages are not supported.")
        dictionary: Dict = {}
        return dictionary

    @_("KEYSTORE_TO_NV mem_opt address_or_range")  # type: ignore
    def keystore_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content keystore statement.
        """
        dictionary = {token.KEYSTORE_TO_NV: {}}
        dictionary[token.KEYSTORE_TO_NV].update(token.mem_opt)
        dictionary[token.KEYSTORE_TO_NV].update(token.address_or_range)
        return dictionary

    @_("KEYSTORE_FROM_NV mem_opt address_or_range")  # type: ignore
    def keystore_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content keystore statement.
        """
        dictionary = {token.KEYSTORE_FROM_NV: {}}
        dictionary[token.KEYSTORE_FROM_NV].update(token.mem_opt)
        dictionary[token.KEYSTORE_FROM_NV].update(token.address_or_range)
        return dictionary

    @_("IDENT")  # type: ignore
    def mem_opt(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        Unsupported syntax right now.

        :param token: object holding the content defined in decorator.
        """
        # search in variables for token.IDENT variable and get it's value
        return {"mem_opt": token.IDENT}

    @_("'@' int_const_expr")  # type: ignore
    def mem_opt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of memory type.
        """
        dictionary = {"mem_opt": token.int_const_expr}
        return dictionary

    @_("empty")  # type: ignore
    def mem_opt(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        Unsupported syntax right now.

        :param token: object holding the content defined in decorator.
        """
        return token.empty

    @_("VERSION_CHECK sec_or_nsec fw_version")  # type: ignore
    def version_stmt(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of version check statement.
        """
        dictionary: Dict = {token.VERSION_CHECK: {}}
        dictionary[token.VERSION_CHECK].update(token.sec_or_nsec)
        dictionary[token.VERSION_CHECK].update(token.fw_version)
        return dictionary

    @_("SEC")  # type: ignore
    def sec_or_nsec(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of version check type.
        """
        dictionary = {"ver_type": 0}
        return dictionary

    @_("NSEC")  # type: ignore
    def sec_or_nsec(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of version check type.
        """
        dictionary = {"ver_type": 1}
        return dictionary

    @_("int_const_expr")  # type: ignore
    def fw_version(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: dictionary holding the content of fw version.
        """
        dictionary = {"fw_version": token.int_const_expr}
        return dictionary

    @_("IF bool_expr LBRACE statement RBRACE else_stmt")  # type: ignore
    def if_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: list of if statements.
        """
        self.error(token, ": if & if-else statement is not supported.")
        if token.bool_expr:
            return token.statement

        return token.else_stmt

    @_("ELSE LBRACE statement RBRACE")  # type: ignore
    def else_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list of else statements.
        """
        return token.statement

    @_("ELSE if_stmt")  # type: ignore
    def else_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: list of else if statements.
        """
        return token.if_stmt

    @_("empty")  # type: ignore
    def else_stmt(self, token: YaccProduction) -> List:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: empty list.
        """
        list_ = []
        return list_

    @_("STRING_LITERAL")  # type: ignore
    def const_expr(self, token: YaccProduction) -> str:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: string.
        """
        return token.STRING_LITERAL[1:-1]

    @_("bool_expr")  # type: ignore
    def const_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: boolean value as a result of constant expression.
        """
        return token.bool_expr

    @_("expr")  # type: ignore
    def int_const_expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: Number as a result of an expression.
        """
        return token.expr

    @_("DEFINED LPAREN IDENT RPAREN")  # type: ignore
    def bool_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: boolean value as a result if some identifier is defined.
        """
        return token.IDENT in self._variables

    @_(  # type: ignore
        "bool_expr LT bool_expr",
        "bool_expr LE bool_expr",
        "bool_expr GT bool_expr",
        "bool_expr GE bool_expr",
        "bool_expr EQ bool_expr",
        "bool_expr NE bool_expr",
        "bool_expr LAND bool_expr",
        "bool_expr LOR bool_expr",
        "LPAREN bool_expr RPAREN",
    )
    def bool_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: boolean value as a result of boolean expression.
        """
        operator = token[1]
        if operator == "<":
            return token.bool_expr0 < token.bool_expr1
        if operator == "<=":
            return token.bool_expr0 <= token.bool_expr1
        if operator == ">":
            return token.bool_expr0 > token.bool_expr1
        if operator == ">=":
            return token.bool_expr0 >= token.bool_expr1
        if operator == "==":
            return token.bool_expr0 == token.bool_expr1
        if operator == "!=":
            return token.bool_expr0 != token.bool_expr1
        if operator == "&&":
            return token.bool_expr0 and token.bool_expr1
        if operator == "||":
            return token.bool_expr0 or token.bool_expr1

        return token[1]

    @_("int_const_expr")  # type: ignore
    def bool_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: boolean value as a result of a boolean expression.
        """
        return token.int_const_expr

    @_("LNOT bool_expr")  # type: ignore
    def bool_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: boolean value as a result of logical not expression.
        """
        return not token.bool_expr

    @_("IDENT LPAREN SOURCE_NAME RPAREN")  # type: ignore
    def bool_expr(self, token: YaccProduction) -> bool:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        :return: boolean value (at the moment always False, as not supported).
        """
        # I've absolutely no clue, what this rule can mean or be for???
        self.error(token, ": IDENT ( SOURCE_NAME ) is not supported.")
        return False

    @_(  # type: ignore
        "expr PLUS expr",
        "expr MINUS expr",
        "expr TIMES expr",
        "expr DIVIDE expr",
        "expr MOD expr",
        "expr LSHIFT expr",
        "expr RSHIFT expr",
        "expr AND expr",
        "expr OR expr",
        "expr XOR expr",
        "expr PERIOD INT_SIZE",
        "LPAREN expr RPAREN",
    )
    def expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: number as a result of an expression.
        """
        operator = token[1]
        if operator == "+":
            return token.expr0 + token.expr1
        if operator == "-":
            return token.expr0 - token.expr1
        if operator == "*":
            return token.expr0 - token.expr1
        if operator == "/":
            return token.expr0 // token.expr1
        if operator == "%":
            return token.expr0 % token.expr1
        if operator == "<<":
            return token.expr0 << token.expr1
        if operator == ">>":
            return token.expr0 >> token.expr1
        if operator == "&":
            return token.expr0 & token.expr1
        if operator == "|":
            return token.expr0 | token.expr1
        if operator == "^":
            return token.expr0 ^ token.expr1
        if operator == ".":
            # TODO maybe we need to store the variable type somewhere so in further
            # operations the number gets truncated on desired size???
            char = token.INT_SIZE
            if char == "w":
                return token[0] & 0xFFFF
            if char == "h":
                return token[0] & 0xFF
            if char == "b":
                return token[0] & 0xF
        # LPAREN expr RPAREN
        return token[1]

    @_("INT_LITERAL")  # type: ignore
    def expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: integer number as a terminal.
        """
        return token.INT_LITERAL

    @_("IDENT")  # type: ignore
    def expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: number stored under identifier.
        """
        # we need to convert the IDENT into a value stored under that identifier
        # search the variables and check, whether there is a name of IDENT
        for var in self._variables:
            if var.name == token.IDENT:
                return var.value

        return token.IDENT

    @_("symbol_ref")  # type: ignore
    def expr(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        """
        self.error(token, ": symbol reference is not supported.")

    @_("unary_expr")  # type: ignore
    def expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: number as a result of unary expression.
        """
        return token.unary_expr

    @_("SIZEOF LPAREN symbol_ref RPAREN")  # type: ignore
    def expr(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        """
        self.error(token, ": sizeof operator is not supported")

    @_("SIZEOF LPAREN IDENT RPAREN")  # type: ignore
    def expr(self, token: YaccProduction) -> None:  # type: ignore
        """Parser rule.

        We don't support this rule for now.

        :param token: object holding the content defined in decorator.
        """
        self.error(token, ": sizeof operator is not supported")

    # OBSOLETE design???
    # @_("IDENT")
    # def source_name(self, token: YaccProduction) -> Optional[str]:
    #     """Parser rule.

    #     :token: object holding the content defined in decorator.
    #     """
    #     # TODO the SOURCE_NAME token seems to be an identifier from the sources
    #     # block defined either as absolute or relative path or with the use of
    #     # extern(number) expression.
    #     # From this perspective, SOURCE_NAME is nothing else then an identifier,
    #     # but with a special meaning. Like a variable of type source...
    #     for v in self._sources:
    #         if v.name == token.IDENT:
    #             return v.value

    #     self.error(token, f": undefined source name '{token.IDENT}'")

    @_("PLUS expr", "MINUS expr")  # type: ignore
    def unary_expr(self, token: YaccProduction) -> Number:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: number as a result of unary expression.
        """
        sign = token[0]
        number = token.expr
        if sign == "-":
            number = -number

        return number

    @_("")  # type: ignore
    def empty(self, token: YaccProduction) -> Dict:  # type: ignore
        """Parser rule.

        :param token: object holding the content defined in decorator.
        :return: empty dictionary.
        """
        dictionary: Dict = {}
        return dictionary

    @staticmethod
    def _find_column(text: str, token: YaccProduction) -> int:
        """Finds the column of token in input.

        :param text: input file being parsed
        :param token: object holding the content defined in decorator.
        :return: column based on token index.
        """
        last_cr = text.rfind("\n", 0, token.index)
        if last_cr < 0:
            last_cr = 0
        else:
            last_cr += 1
        column = (token.index - last_cr) + 1
        return column

    @staticmethod
    def _find_line(text: str, line_num: int) -> str:
        """Finds the line in text based on line number.

        :param text: text to return required line.
        :param line_num: line number to return.
        :return: line 'line_num" in 'text'.
        """
        lines = text.split("\n")

        return lines[line_num]

    def error(
        self, token: YaccProduction, msg: str = ""
    ) -> YaccProduction:  # pylint: disable=arguments-differ
        """Syntax error handler.

        On syntax error, we set an error flag and read the rest of input file
        until end to terminate the process of parsing.

        :param token: object holding the content defined in decorator.
        :param msg: error message to use.

        :raises SPSDKError: Raises error with 'msg' message.
        """
        self._parse_error = True

        if token:
            lineno = getattr(token, "lineno", -1)
            if lineno != -1:
                column = BDParser._find_column(self._input, token)
                error_line = BDParser._find_line(self._input, lineno - 1)
                raise SPSDKError(
                    f"bdcompiler:{lineno}:{column}: error{msg}\n\n{error_line}\n"
                    + (column - 1) * " "
                    + "^\n"
                )

            raise SPSDKError(f"bdcompiler: error{msg}\n")

        raise SPSDKError("bdcompiler: unspecified error.")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module implementing the TrustProvisioning Data Container."""

import struct
from abc import abstractmethod
from typing import Any, List, Mapping, Type

import hexdump

from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import align, align_block

from ..exceptions import SPSDKTpError
from .data_container_auth import AuthenticationType, get_auth_data_len, sign, validate
from .payload_types import PayloadType

ALIGNMENT = 8


class EntryType(Enum):
    """Enumeration of all supported Entry types."""

    STANDARD = (0xA0, "standard", "Standard Entry")
    DESTINATION = (0xB0, "destination", "Destination Entry")
    AUTHENTICATION = (0xC0, "authentication", "Authentication Entry")


class DestinationType(Enum):
    """Destination type setting for DataDestinationEntry."""

    MEMORY = (0, "memory", "Address in memory")
    OTP = (2, "otp", "Index in OTP")


class BaseElement:
    """Base class for items used in data_container."""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {vars(self)}>"

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, self.__class__) and vars(self) == vars(other)

    @abstractmethod
    def export(self) -> bytes:
        """Serialize object data."""

    @classmethod
    @abstractmethod
    def parse(cls, data: bytes, offset: int = 0) -> "BaseElement":
        """Reconstruct object from serialized data."""


class EntryHeader(BaseElement):
    """Common Entry header for all entry types."""

    #: Binary format for entry header (used by the struct module)
    FORMAT = "<H2B2H"
    #: Total size of the entry header
    SIZE = struct.calcsize(FORMAT)

    def __init__(
        self,
        tag: int,
        payload_size: int,
        payload_type: int,
        entry_extra: int = 0,
    ) -> None:
        """Initialize Entry header.

        :param tag: Tag indicating entry type, see `EntryType`
        :param payload_size: Size of entry payload, excluding header and padding
        :param payload_type: Type of entry payload
        :param entry_extra: Extra type specific data
        """
        self.tag = tag
        self.payload_size = payload_size
        self.payload_type = payload_type
        self.entry_extra = entry_extra

    def stringify_payload_type(self) -> str:
        """Return a stringified payload type."""
        enum = AuthenticationType if self.tag == EntryType.AUTHENTICATION else PayloadType
        return f"{self.payload_type:#06x} - {enum.desc(self.payload_type)}"  # type: ignore

    def __str__(self) -> str:
        info = (
            f"Entry type:   {self.tag:#x} - {EntryType.desc(self.tag)}\n"
            f"Entry size:   {self.payload_size:#06x} - {self.payload_size}\n"
            f"Entry extra:  {self.entry_extra:#06x} - {self.entry_extra}\n"
        )
        return info

    def export(self) -> bytes:
        """Serialize the entry header."""
        data = struct.pack(
            self.FORMAT,
            self.payload_size,
            0,
            self.tag,
            self.entry_extra,
            self.payload_type,
        )
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "EntryHeader":
        """Reconstruct the entry header from binary data."""
        size, _, tag, extra, p_type = struct.unpack_from(cls.FORMAT, data, offset)
        return cls(tag=tag, payload_size=size, payload_type=p_type, entry_extra=extra)


class DestinationHeader(BaseElement):
    """Header used to store destination information."""

    #: Binary format for destination record (used by the struct module)
    FORMAT = "<4BL"
    #: Total size of the destination record
    SIZE = struct.calcsize(FORMAT)

    def __init__(self, destination: int, destination_type: DestinationType) -> None:
        """Initialize the destination record.

        :param destination: Destination address
        :param destination_type: Destination type, see: `DestinationType`
        """
        self.destination = destination
        self.destination_type = destination_type

    def __str__(self) -> str:
        info = (
            f"Dest. type:   {DestinationType.desc(self.destination_type)}\n"
            f"Destination:  {self.destination:#010x}\n"
        )
        return info

    def export(self) -> bytes:
        """Serialize the destination record."""
        data = struct.pack(self.FORMAT, self.destination_type, 0, 0, 0, self.destination)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DestinationHeader":
        """Reconstruct the destination record from binary data."""
        dest_type, _, _, _, dest = struct.unpack_from(cls.FORMAT, data, offset)
        return cls(destination=dest, destination_type=dest_type)


class DataEntry(BaseElement):
    """Standard data entry."""

    TAG = EntryType.STANDARD

    def __init__(self, payload: bytes, payload_type: int, extra: int = 0) -> None:
        """Initialize Standard Data Entry.

        :param payload: Data entry payload
        :param payload_type: Data entry type
        :param extra: Extra information for the entry
        """
        self.payload = payload
        self.header = EntryHeader(
            tag=self.TAG, payload_size=len(payload), payload_type=payload_type, entry_extra=extra  # type: ignore
        )

    def _stringify_payload(self) -> str:
        info = f"Payload type: {self.header.stringify_payload_type()}\n"
        info += "Payload data: "
        if len(self.payload) <= 4:
            info += f"{self.payload.hex()}\n"
        else:
            info += f"\n{hexdump.hexdump(self.payload, result='return')}\n"
        return info

    def __str__(self) -> str:
        info = str(self.header)
        info += self._stringify_payload()
        return info

    @property
    def total_size(self) -> int:
        """Returns total size of the entry (including header and padding)."""
        size = struct.calcsize(self.header.FORMAT)
        size += align(len(self.payload), alignment=ALIGNMENT)
        return size

    def export(self) -> bytes:
        """Serialize the entry."""
        data = self.header.export()
        data += self.payload
        return align_block(data, alignment=ALIGNMENT)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DataEntry":
        """Reconstruct the entry from binary data."""
        header = EntryHeader.parse(data, offset=offset)
        payload_offset = header.SIZE + offset
        payload_length = header.payload_size
        payload = data[payload_offset : payload_offset + payload_length]
        return cls(payload=payload, payload_type=header.payload_type, extra=header.entry_extra)


class DataDestinationEntry(DataEntry):
    """Data entry including destination information."""

    TAG = EntryType.DESTINATION

    def __init__(
        self,
        payload: bytes,
        payload_type: int,
        destination: int,
        destination_type: DestinationType,
        extra: int = 0,
    ) -> None:
        """Initialize data entry with destination record.

        :param payload: Data entry payload
        :param payload_type: Data entry type
        :param destination: Destination memory address/OTP index
        :param destination_type: Destination type memory/OTP
        """
        super().__init__(payload, payload_type, extra)
        self.destination_header = DestinationHeader(
            destination=destination, destination_type=destination_type
        )

    def __str__(self) -> str:
        info = str(self.header)
        info += str(self.destination_header)
        info += self._stringify_payload()
        return info

    @property
    def total_size(self) -> int:
        """Returns total size of the entry (including header and padding)."""
        size = super().total_size
        size += struct.calcsize(self.destination_header.FORMAT)
        return size

    def export(self) -> bytes:
        """Serialize the entry."""
        data = self.header.export()
        data += self.destination_header.export()
        data += self.payload
        return align_block(data, alignment=ALIGNMENT)

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "DataDestinationEntry":
        """Reconstruct the entry header from binary data."""
        header = EntryHeader.parse(data, offset=offset)
        dest_header = DestinationHeader.parse(data, offset=offset + header.SIZE)
        payload_start = offset + header.SIZE + dest_header.SIZE
        payload_size = header.payload_size
        payload = data[payload_start : payload_start + payload_size]
        return cls(
            payload=payload,
            payload_type=header.payload_type,
            extra=header.entry_extra,
            destination=dest_header.destination,
            destination_type=dest_header.destination_type,
        )


class DataAuthenticationEntry(DataEntry):
    """Final Data entry used for integrity check."""

    TAG = EntryType.AUTHENTICATION


class ContainerHeader(BaseElement):
    """Main container header."""

    #: Binary format for entry header (used by the struct module)
    FORMAT = "<4B2H"
    #: Total size of the container header
    SIZE = struct.calcsize(FORMAT)
    #: Tag (magic) indicating start of the container in binary stream
    TAG = 0x33

    def __init__(self, major: int = 1, minor: int = 0, patch: int = 0) -> None:
        """Initialize Container Header with version information.

        :param major: major version, defaults to 1
        :param minor: minor version, defaults to 0
        :param patch: patch re-vision, defaults to 0
        """
        self.size = 0
        self.major = major
        self.minor = minor
        self.patch = patch

    def __str__(self) -> str:
        info = f"Version: {self.major}.{self.minor}.{self.patch}\n"
        info += f"Size:    {self.size:#x} - {self.size}\n"
        return info

    def export(self) -> bytes:
        """Serialize the container header."""
        data = struct.pack(self.FORMAT, self.patch, self.minor, self.major, self.TAG, self.size, 0)
        return data

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "ContainerHeader":
        """Reconstruct the container header from binary data."""
        patch, minor, major, tag, size, _ = struct.unpack_from(cls.FORMAT, data, offset)
        if tag != cls.TAG:
            raise SPSDKTpError(f"Invalid TAG found: {hex(tag)}, expected {hex(cls.TAG)}")
        header = ContainerHeader(major=major, minor=minor, patch=patch)
        header.size = size
        return header


class Container(BaseElement):
    """TrustProvisioning Data Container."""

    def __init__(self) -> None:
        """Initialize the container."""
        self.header = ContainerHeader()
        #: list containing individual entries (shall not be modified directly)
        self._entries: List[DataEntry] = []

    def __str__(self) -> str:
        info = (
            f"TrustProvisioning Data Container\n"
            f"{self.header}"
            f"Entries: (count: {len(self._entries)})\n"
        )
        for entry in self._entries:
            info += "-" * 76 + "\n"
            info += str(entry)
        return info

    def __len__(self) -> int:
        return len(self._entries)

    def add_entry(self, entry: DataEntry) -> None:
        """Add an entry to the container."""
        self._entries.append(entry)
        self._update_header()

    def _update_header(self) -> None:
        """Update size information in container header."""
        size = sum(entry.total_size for entry in self._entries)
        self.header.size = size

    def export(self) -> bytes:
        """Serialize the container."""
        data = self.header.export()
        for entry in self._entries:
            data += entry.export()
        return data

    def get_entry(self, payload_type: PayloadType) -> DataEntry:
        """Get the first entry for the given type.

        :param payload_type: Type of an entry to look for
        :raises SPSDKTpError: Container doesn't have an entry of required type
        :return: Entry with given type
        """
        for entry in self._entries:
            if entry.header.payload_type == payload_type:
                return entry
        raise SPSDKTpError(
            f"Container doesn't have an entry with type {PayloadType.name(payload_type)}"
        )

    def get_entries(self, payload_type: PayloadType) -> List[DataEntry]:
        """Get all entries for given payload type."""
        return [entry for entry in self._entries if entry.header.payload_type == payload_type]

    def add_auth_entry(self, auth_type: AuthenticationType, key: bytes) -> None:
        """Add the final data authentication entry.

        :param auth_type: Authentication Type
        :param key: Key for authentication
        :raises SPSDKTpError: Unknown or not-implemented Authentication type
        """
        if any(isinstance(entry, DataAuthenticationEntry) for entry in self._entries):
            raise SPSDKTpError("Can't add additional AuthEntry")
        # to add DataAuth entry we need all data so far + header of a not-yet-existing entry :/
        # on top of that we need to add "pro-forma" entry to update container header
        auth_data_len = get_auth_data_len(auth_type=auth_type)
        self.add_entry(
            DataAuthenticationEntry(payload=bytes(auth_data_len), payload_type=auth_type)
        )
        data = self.export()

        # find actual data to sign (skip the pro-forma signature)
        auth_entry = self._entries[-1]
        raw_data_end = auth_entry.total_size
        raw_data_end -= EntryHeader.SIZE
        data_to_sign = data[:-raw_data_end]

        signature = sign(data=data_to_sign, auth_type=auth_type, key=key)
        # update payload for last entry (signature)
        self._entries[-1].payload = signature

    def validate(self, key: bytes) -> bool:
        """Validate signature/authentication code.

        :param key: Key for validating signature
        :raises SPSDKTpError: Unknown or non-implemented Authentication type
        :return: True if signature/authentication code is valid
        """
        if not isinstance(self._entries[-1], DataAuthenticationEntry):
            raise SPSDKTpError("DataAuth Entry is not at the end of container")
        data = self.export()
        # to perform validation we need to split serialized:
        # - 'raw_data', signature, and padding
        # raw_data starts at the begging and ends with header of last entry (including the header)
        auth_entry = self._entries[-1]
        raw_data_end = auth_entry.total_size
        raw_data_end -= EntryHeader.SIZE
        data_to_validate = data[:-raw_data_end]

        return validate(
            data=data_to_validate,
            signature=auth_entry.payload,
            auth_type=auth_entry.header.payload_type,  # type: ignore
            key=key,
        )

    @classmethod
    def parse(cls, data: bytes, offset: int = 0) -> "Container":
        """Reconstruct container from binary data."""
        header = ContainerHeader.parse(data=data, offset=offset)
        offset += ContainerHeader.SIZE
        container = cls()
        while offset < len(data) and offset < header.size:
            entry = parse_entry(data=data, offset=offset)
            offset += entry.total_size
            container.add_entry(entry=entry)
        return container


#: Mapping between entry type and its corresponding DataEntry class
_ENTRY_CLASSES: Mapping[int, Type[DataEntry]] = {
    EntryType.STANDARD: DataEntry,  # type: ignore
    EntryType.DESTINATION: DataDestinationEntry,  # type: ignore
    EntryType.AUTHENTICATION: DataAuthenticationEntry,  # type: ignore
}


def parse_entry(data: bytes, offset: int = 0) -> "DataEntry":
    """Common parser for all known DataEntry classes."""
    tag = data[offset + 3]
    return _ENTRY_CLASSES[tag].parse(data=data, offset=offset)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module defining a USB filtering class."""
import platform
import re
from typing import Any, Dict, Optional, Tuple

from .misc import get_hash


class USBDeviceFilter:
    """Generic USB Device Filtering class.

    Create a filtering instance. This instance holds the USB ID you are interested
    in during USB HID device search and allows you to compare, whether
    provided USB HID object is the one you are interested in.
    The allowed format of `usb_id` string is following:

    vid or pid - vendor ID or product ID. String holding hex or dec number.
    Hex number must be preceded by 0x or 0X. Number of characters after 0x is
    1 - 4. Mixed upper & lower case letters is allowed. e.g. "0xaB12", "0XAB12",
    "0x1", "0x0001".
    The decimal number is restricted only to have 1 - 5 digits, e.g. "65535"
    It's allowed to set the USB filter ID to decimal number "99999", however, as
    the USB VID number is four-byte hex number (max value is 65535), this will
    lead to zero results. Leading zeros are not allowed e.g. 0001. This will
    result as invalid match.

    The user may provide a single number as usb_id. In such a case the number
    may represent either VID or PID. By default, the filter expects this number
    to be a VID. In rare cases the user may want to filter based on PID.
    Initialize the `search_by_pid` parameter to True in such cases.

    vid/pid - string of vendor ID & product ID separated by ':' or ','
    Same rules apply to the number format as in VID case, except, that the
    string consists of two numbers separated by ':' or ','. It's not allowed
    to mix hex and dec numbers, e.g. "0xab12:12345" is not allowed.
    Valid vid/pid strings:
    "0x12aB:0xabc", "1,99999"

    Windows specific:
    instance ID - String in following format "HID\\VID_<HEX>&PID_<HEX>\\<instance_id>",
    see instance ID in device manager under Windows OS.

    Linux specific:
    USB device path - HID API returns path in following form:
    '0003:0002:00'

    The first number represents the Bus, the second Device and the third interface. The Bus:Device
    number is unique so interface is not necessary and Bus:Device should be sufficient.

    The Bus:Device can be observed using 'lsusb' command. The interface can be observed using
    'lsusb -t'. lsusb returns the Bus and Device as a 3-digit number.
    It has been agreed, that the expected input is:
    <Bus in dec>#<Device in dec>, e.g. 3#11

    Mac specific:
    USB device path - HID API returns path in roughly following form:
    'IOService:/AppleACPIPlatformExpert/PCI0@0/AppleACPIPCI/XHC1@14/XHC1@14000000/HS01@14100000/SE
    Blank RT Family @14100000/IOUSBHostInterface@0/AppleUserUSBHostHIDDevice'

    This path can be found using the 'ioreg' utility or using 'IO Hardware Registry Explorer' tool.
    However, using the system report from 'About This MAC -> System Report -> USB' a partial path
    can also be gathered. Using the name of USB device from the 'USB Device Tree' and appending
    the 'Location ID' should work. The name can be 'SE Blank RT Family' and the 'Location ID' is
    in form <hex> / <dec>, e.g. '0x14200000 / 18'.
    So the 'usb_id' name should be 'SE Blank RT Family @14200000' and the filter should be able to
    filter out such device.
    """

    def __init__(
        self,
        usb_id: Optional[str] = None,
        search_by_pid: bool = False,
    ):
        """Initialize the USB Device Filtering.

        :param usb_id: usb_id string
        :param search_by_pid: if true, expects usb_id to be a PID number, VID otherwise.
        """
        self.usb_id = usb_id
        self.search_by_pid = search_by_pid

    def compare(self, usb_device_object: Dict[str, Any]) -> bool:
        """Compares the internal `usb_id` with provided `usb_device_object`.

        The provided USB ID during initialization may be VID or PID, VID/PID pair,
        or a path. See private methods for details.

        :param usb_device_object: Libusbsio/HID_API device object (dictionary)

        :return: True on match, False otherwise
        """
        # Determine, whether given device matches one of the expected criterion
        if self.usb_id is None:
            return True

        vendor_id = usb_device_object.get("vendor_id")
        product_id = usb_device_object.get("product_id")
        serial_number = usb_device_object.get("serial_number")
        device_name = usb_device_object.get("device_name")
        # the Libusbsio/HID_API holds the path as bytes, so we convert it to string
        usb_path_raw = usb_device_object.get("path")

        if usb_path_raw:
            if self.usb_id == get_hash(usb_path_raw):
                return True
            usb_path = self.convert_usb_path(usb_path_raw)
            if self._is_path(usb_path=usb_path):
                return True

        if self._is_vid_or_pid(vid=vendor_id, pid=product_id):
            return True

        if vendor_id and product_id and self._is_vid_pid(vid=vendor_id, pid=product_id):
            return True

        if serial_number and self.usb_id.casefold() == serial_number.casefold():
            return True

        if device_name and self.usb_id.casefold() == device_name.casefold():
            return True

        return False

    def _is_path(self, usb_path: str) -> bool:
        """Compares the internal usb_id with provided path.

        If the path is a substring of the usb_id, this is considered as a match
        and True is returned.

        :param usb_path: path to be compared with usd_id.
        :return: true on a match, false otherwise.
        """
        # we check the len of usb_id, because usb_id = "" is considered
        # to be always in the string returning True, which is not expected
        # behavior
        # the provided usb string id fully matches the instance ID
        usb_id = self.usb_id or ""
        if usb_id.casefold() in usb_path.casefold() and len(usb_id) > 0:
            return True

        return False

    def _is_vid_or_pid(self, vid: Optional[int], pid: Optional[int]) -> bool:
        # match anything starting with 0x or 0X followed by 0-9 or a-f or
        # match either 0 or decimal number not starting with zero
        # this regex is the same for vid and pid => xid
        xid_regex = "0[xX][0-9a-fA-F]{1,4}|0|[1-9][0-9]{0,4}"
        usb_id = self.usb_id or ""
        if re.fullmatch(xid_regex, usb_id) is not None:
            # the string corresponds to the vid/pid specification, check a match
            if self.search_by_pid and pid:
                if int(usb_id, 0) == pid:
                    return True
            elif vid:
                if int(usb_id, 0) == vid:
                    return True

        return False

    def _is_vid_pid(self, vid: int, pid: int) -> bool:
        """If usb_id corresponds to VID/PID pair, compares it with provided vid/pid.

        :param vid: vendor ID to compare.
        :param pid: product ID to compare.
        :return: true on a match, false otherwise.
        """
        # match anything starting with 0x or 0X followed by 0-9 or a-f or
        # match either 0 or decimal number not starting with zero
        # Above pattern is combined to match a pair corresponding to vid/pid.
        vid_pid_regex = "0[xX][0-9a-fA-F]{1,4}(,|:)0[xX][0-9a-fA-F]{1,4}|(0|[1-9][0-9]{0,4})(,|:)(0|[1-9][0-9]{0,4})"
        usb_id = self.usb_id or ""
        if re.fullmatch(vid_pid_regex, usb_id):
            # the string corresponds to the vid/pid specification, check a match
            vid_pid = re.split(":|,", usb_id)
            if vid == int(vid_pid[0], 0) and pid == int(vid_pid[1], 0):
                return True

        return False

    @staticmethod
    def convert_usb_path(hid_api_usb_path: bytes) -> str:
        """Converts the Libusbsio/HID_API path into string, which can be observed from OS.

        DESIGN REMARK: this function is not part of the USBLogicalDevice, as the
        class intention is to be just a simple container. But to help the class
        to get the required inputs, this helper method has been provided. Additionally,
        this method relies on the fact that the provided path comes from the Libusbsio/HID_API.
        This method will most probably fail or provide improper results in case
        path from different USB API is provided.

        :param hid_api_usb_path: USB device path from Libusbsio/HID_API
        :return: Libusbsio/HID_API path converted for given platform
        """
        if platform.system() == "Windows":
            device_manager_path = hid_api_usb_path.decode("utf-8").upper()
            device_manager_path = device_manager_path.replace("#", "\\")
            result = re.search(r"\\\\\?\\(.+?)\\{", device_manager_path)
            if result:
                device_manager_path = result.group(1)

            return device_manager_path

        if platform.system() == "Linux":
            # we expect the path in form of <bus>#<device>, Libusbsio/HID_API returns
            # <bus>:<device>:<interface>
            linux_path = hid_api_usb_path.decode("utf-8")
            linux_path_parts = linux_path.split(":")

            if len(linux_path_parts) > 1:
                linux_path = str.format(
                    "{}#{}", int(linux_path_parts[0], 16), int(linux_path_parts[1], 16)
                )

            return linux_path

        if platform.system() == "Darwin":
            return hid_api_usb_path.decode("utf-8")

        return ""


class NXPUSBDeviceFilter(USBDeviceFilter):
    """NXP Device Filtering class.

    Extension of the generic USB device filter class to support filtering
    based on NXP devices. Modifies the way, how single number is handled.
    By default, if single value is provided, it's content is expected to be VID.
    However, legacy tooling were expecting PID, so from this perspective if
    a single number is provided, we expect that VID is out of range NXP_VIDS.
    """

    NXP_VIDS = [0x1FC9, 0x15A2, 0x0471, 0x0D28]

    def __init__(
        self,
        usb_id: Optional[str] = None,
        nxp_device_names: Optional[Dict[str, Tuple[int, int]]] = None,
    ):
        """Initialize the USB Device Filtering.

        :param usb_id: usb_id string
        :param nxp_device_names: Dictionary holding NXP device vid/pid {"device_name": [vid(int), pid(int)]}
        """
        super().__init__(usb_id=usb_id, search_by_pid=True)
        self.nxp_device_names = nxp_device_names or {}

    def compare(self, usb_device_object: Any) -> bool:
        """Compares the internal `usb_id` with provided `usb_device_object`.

        Extends the comparison by USB names - dictionary of device name and
        corresponding VID/PID.

        :param usb_device_object: lpcusbsio USB HID device object

        :return: True on match, False otherwise
        """
        vendor_id = usb_device_object["vendor_id"]
        product_id = usb_device_object["product_id"]

        if self.usb_id:
            if super().compare(usb_device_object=usb_device_object):
                return True

            return self._is_nxp_device_name(vendor_id, product_id)

        return self._is_nxp_device(vendor_id)

    def _is_vid_or_pid(self, vid: Optional[int], pid: Optional[int]) -> bool:
        if vid and vid in NXPUSBDeviceFilter.NXP_VIDS:
            return super()._is_vid_or_pid(vid, pid)

        return False

    def _is_nxp_device_name(self, vid: int, pid: int) -> bool:
        nxp_device_name_to_compare = {k.lower(): v for k, v in self.nxp_device_names.items()}
        assert isinstance(self.usb_id, str)
        if self.usb_id.lower() in nxp_device_name_to_compare:
            vendor_id, product_id = nxp_device_name_to_compare[self.usb_id.lower()]
            if vendor_id == vid and product_id == pid:
                return True
        return False

    @staticmethod
    def _is_nxp_device(vid: int) -> bool:
        return vid in NXPUSBDeviceFilter.NXP_VIDS



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Trust provisioning - TP Device, Smart Card Adapter."""

import logging
import os
from enum import Enum
from typing import Any, Dict, List, Optional, cast

from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.serialization import Encoding
from cryptography.x509 import NameOID, load_der_x509_certificate

from spsdk import SPSDKError
from spsdk.crypto.certificate_management import (
    convert_certificate_into_bytes,
    generate_certificate,
    generate_name,
)
from spsdk.crypto.keys_management import generate_ecc_public_key
from spsdk.crypto.loaders import extract_public_key, load_private_key
from spsdk.utils.database import Database
from spsdk.utils.misc import (
    find_file,
    get_hash,
    load_binary,
    load_text,
    numberify_version,
    value_to_int,
)
from spsdk.utils.schema_validator import ValidationSchemas

try:
    from smartcard.CardConnectionDecorator import CardConnection
except ImportError as e:
    raise SPSDKError(
        "pyscard package is missing, please install it with pip install 'spsdk[tp]' in order to use TP"
    ) from e


from .. import TP_DATA_FOLDER, TP_SCH_FILE, SPSDKTpError, TpDevInterface
from ..tp_intf import TpIntfDescription
from . import scard_commands
from .scard_utils import ProvItem, get_applet_infos
from .utils import sanitize_common_name

logger = logging.getLogger(__name__)
TP_DATA_FILE = os.path.join(TP_DATA_FOLDER, "database.yaml")


class TpSCardDescription(TpIntfDescription):
    """Smart Card interface description."""

    def __init__(
        self,
        name: str,
        version: str,
        serial_number: int,
        settings: Optional[Dict] = None,
        card_connection: Optional[CardConnection] = None,
    ) -> None:
        """Smart Card interface description."""
        super().__init__(name, TpDevSmartCard, "TP HSM Applet", settings, version)
        self.serial_number = serial_number
        self.card_connection = card_connection

    def get_id(self) -> int:
        """Returns the ID of the interface (smart card serial number)."""
        return self.serial_number

    def get_id_hash(self) -> str:
        """Return the ID hash of the interface. (hash of the reader's name)."""
        return get_hash(self.name)

    def as_dict(self) -> Dict[str, Any]:
        """Returns whole record as dictionary suitable for printing."""
        data = super().as_dict()
        # don't display card_connection object in list
        data.pop("card_connection")
        data["name_hash"] = self.get_id_hash()
        return data


class TpDevSmartCard(TpDevInterface):
    """Trust provisioning device adapter for Smart Card."""

    NAME = "scard"
    ATR = "3BD518FF8191FE1FC38073C821100A"
    APPLET = "tphsmapplet"
    MIN_VERSION = "1.0.6"

    class SettingsKey(str, Enum):
        """Keys used in `get_connected_devices` in `settings` dictionary."""

        ID = "id"
        READER = "reader"

    @classmethod
    def get_connected_devices(cls, settings: Optional[Dict] = None) -> List[TpIntfDescription]:
        """Get all connected TP devices of this adapter.

        :param settings: Possible settings to determine the way to find connected device, defaults to None.
        :return: List of all founded TP devices.
        """
        ret: List[TpIntfDescription] = []
        filter_id = None if settings is None else settings.get("id")
        reader = None if settings is None else settings.get("reader")
        cards = get_applet_infos(
            atr=cls.ATR, applet=cls.APPLET, filter_id=filter_id, filter_reader=reader
        )
        ret = [
            TpSCardDescription(
                name=card.reader_name,
                version=card.version,
                serial_number=card.serial_number,
                settings=settings,
                card_connection=card.card_connection,
            )
            for card in cards
        ]
        return ret

    get_connected_interfaces = get_connected_devices

    def __init__(self, descriptor: TpIntfDescription) -> None:
        """Initialize Smart Card Adapter."""
        if not descriptor.name or not isinstance(descriptor, TpSCardDescription):
            raise SPSDKTpError("Invalid SMARTCard Interface descriptor.")
        super().__init__(descriptor=descriptor)

        self.descriptor = cast(TpSCardDescription, self.descriptor)
        self.card_connection: Optional[CardConnection] = None

    def open(self) -> None:
        """Open the TP device adapter."""
        # MyPy doesn't pick up it's own "cast" operator :/
        assert isinstance(self.descriptor, TpSCardDescription)

        filter_id = None if self.descriptor.settings is None else self.descriptor.settings.get("id")
        if self.descriptor.card_connection:
            self.card_connection = self.descriptor.card_connection
        else:
            cards = get_applet_infos(atr=self.ATR, applet=self.APPLET, filter_id=filter_id)
            if not cards:
                raise SPSDKTpError("Smart card is not available.")
            if len(cards) != 1:
                raise SPSDKTpError("Multiple suitable cards found.")
            # it's possible tp use descriptor without an connection
            self.card_connection = cards[0].card_connection

        if numberify_version(self.descriptor.version) < numberify_version(self.MIN_VERSION):
            raise SPSDKTpError(
                f"Applet is out-of-date, "
                f"Min required version: {self.MIN_VERSION}, provided: {self.descriptor.version}"
            )
        logger.info(
            f"Connecting to '{self.card_connection.getReader()}'"
            f", applet: '{self.APPLET}'"
            f", serial number: {self.descriptor.serial_number}"
        )
        self.card_connection.connect()
        self.select_applet(self.APPLET)

    def close(self) -> None:
        """Close the TP device adapter."""
        if self.card_connection:
            self.card_connection.disconnect()

    def get_challenge(self, timeout: Optional[int] = None) -> bytes:
        """Request challenge from the TP device.

        :param timeout: Timeout of operation in milliseconds.
        :raises SPSDKTpError: Smart card is not ready.
        :return: Challenge record, to be used for TP communication.
        """
        if not self.card_connection:
            raise SPSDKTpError("Smart card is not ready.")

        logger.debug("Getting challenge")
        challenge_cmd = scard_commands.GetChallenge()
        challenge = challenge_cmd.transmit(self.card_connection)
        logger.debug("Getting challenge complete")
        return challenge

    def authenticate_response(self, tp_data: bytes, timeout: Optional[int] = None) -> bytes:
        """Request TP device for TP authentication of connected MCU.

        :param tp_data: TP response of connected MCU.
        :param timeout: Timeout of operation in milliseconds.
        :raises SPSDKTpError: Smart card is not ready.
        :return: Wrapped data after TP response processing.
        """
        if not self.card_connection:
            raise SPSDKTpError("Smart card is not ready.")

        logger.debug("Authenticating TP_RESPONSE")
        auth_resp_cmd = scard_commands.ProcessTPResponse(tp_response=tp_data)
        wrap_data = auth_resp_cmd.transmit(self.card_connection)
        logger.debug("Authenticating TP_RESPONSE complete")
        return wrap_data

    @staticmethod
    def get_help() -> str:
        """Return help for this interface, including settings description."""
        return "\n".join(
            [
                "The adapter for SMARTCARD readers and cards support.",
                "There are two identification options (at least one is required)",
                "   - id - Serial number of the Smart Card.",
                "   - reader - Name of the Smart Card reader.",
            ]
        )

    @classmethod
    def get_validation_schemas(cls) -> List[Dict[str, Any]]:
        """Return all additional validation schemas for interface.

        return: List of all additional validation schemas.
        """
        sch_cfg_file = ValidationSchemas.get_schema_file(TP_SCH_FILE)

        return [sch_cfg_file["device_scard"]]

    def select_applet(self, applet: str) -> None:
        """Select applet in smart card.

        :param applet: Identification of required Applet.
        :raises SPSDKTpError: Smart card is not ready.
        """
        if not self.card_connection:
            raise SPSDKTpError("Smart card is not ready.")

        logger.debug(f"Selecting smartcard Applet: {applet}")
        select_cmd = scard_commands.Select(applet)
        select_cmd.transmit(self.card_connection)
        logger.debug("Applet selected successfully.")

    def get_applet_version(self) -> str:
        """Get selected applet version from smart card.

        :raises SPSDKTpError: Smart card is not ready.
        :return: Version of selected applet in card.
        """
        if not self.card_connection:
            raise SPSDKTpError("Smart card is not ready.")

        logger.debug("Getting selected applet version.")
        version_cmd = scard_commands.GetAppletVersion()
        applet_version = version_cmd.format(
            version_cmd.transmit(self.card_connection),
        )
        logger.debug(f"Applet version: {applet_version}")
        return applet_version

    def upload(self, config_data: dict, config_dir: Optional[str] = None) -> None:
        """Upload Provisioning data."""
        logger.info("Sending CMPA")
        file_path = config_data["cmpa_path"]
        cmpa = load_binary(file_path, search_paths=[config_dir] if config_dir else None)
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.CMPA, data=cmpa)
        cmd.transmit(self.card_connection)

        logger.info("Sending CFPA")
        file_path = config_data["cfpa_path"]
        cfpa = load_binary(file_path, search_paths=[config_dir] if config_dir else None)
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.CFPA, data=cfpa)
        cmd.transmit(self.card_connection)

        logger.info("Sending SB KEK")
        file_path = config_data["sb_kek_path"]
        sb_kek = bytes.fromhex(
            load_text(file_path, search_paths=[config_dir] if config_dir else None).strip()
        )
        # Send KEK reversed, so Applet doesn't have to reverse it
        sb_kek = bytes(reversed(sb_kek))
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.SB_KEK, data=sb_kek)
        cmd.transmit(self.card_connection)

        logger.info("Sending USER KEK")
        file_path = config_data["user_kek_path"]
        user_kek = bytes.fromhex(
            load_text(file_path, search_paths=[config_dir] if config_dir else None).strip()
        )
        # Send KEK reversed, so Applet doesn't have to reverse it
        user_kek = bytes(reversed(user_kek))
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.USER_KEK, data=user_kek)
        cmd.transmit(self.card_connection)

        logger.info("Sending oem_log_prk")
        file_path = config_data["oem_log_prk_path"]
        log_prk_data = self._serialize_private_key(file_path, [config_dir] if config_dir else None)
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.OEM_LOG_PRK, data=log_prk_data)
        cmd.transmit(self.card_connection)

        logger.info("Sending nxp_global_attest_puk")
        file_path = config_data["nxp_global_attest_puk_path"]
        nxp_glob_puk_file = find_file(file_path, search_paths=[config_dir] if config_dir else None)
        puk = extract_public_key(nxp_glob_puk_file)
        assert isinstance(puk, ec.EllipticCurvePublicKey)
        x = puk.public_numbers().x.to_bytes(length=32, byteorder="big")
        y = puk.public_numbers().y.to_bytes(length=32, byteorder="big")

        cmd = scard_commands.SetProvisioningItem(
            prov_item=ProvItem.NXP_GLOB_PUK, data=b"\x04" + x + y
        )
        cmd.transmit(self.card_connection)

        logger.info("Sending nxp_prod_cert_path")
        file_path = config_data["nxp_prod_cert_path"]
        cert_data = self._serialize_cert(file_path, [config_dir] if config_dir else None)
        cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.NXP_PROD_CERT, data=cert_data)
        cmd.transmit(self.card_connection)

        logger.info("Sending production_quota")
        quota: int = config_data["production_quota"]
        cmd = scard_commands.SetProvisioningItem(
            prov_item=ProvItem.PROD_COUNTER, data=quota.to_bytes(length=4, byteorder="big")
        )
        cmd.transmit(self.card_connection)

        logger.info("Sending oem_id_count")
        oem_id_count: int = config_data["oem_id_count"]
        # TODO: figure out the key flags 1:0; for now we put 00
        oem_key_flags = oem_id_count << 2
        cmd = scard_commands.SetProvisioningItem(
            prov_item=ProvItem.OEM_CERT_COUNT,
            data=oem_key_flags.to_bytes(length=1, byteorder="big"),
        )
        cmd.transmit(self.card_connection)

        if oem_id_count > 0:
            logger.info("Sending oem_id_prk")
            file_path = config_data["oem_id_prk_path"]
            prk_data = self._serialize_private_key(file_path, [config_dir] if config_dir else None)
            cmd = scard_commands.SetProvisioningItem(prov_item=ProvItem.OEM_ID_PRK, data=prk_data)
            cmd.transmit(self.card_connection)

            cert_template = self._create_oem_cert_template(config_data, config_dir)
            for i in range(oem_id_count):
                cert_data = self._serialize_cert_data(
                    cert_template, config_data["oem_id_addresses"][i]
                )
                logger.info(f"Sending OEM cert template #{i+1}")
                cmd = scard_commands.SetProvisioningItem(
                    prov_item=ProvItem.OEM_CERT_TEMPLATE + i, data=cert_data
                )
                cmd.transmit(self.card_connection)

    def setup(self) -> None:
        """Setup the provisioning device."""
        logger.info("Setting up a Smart Card")
        try:
            logger.info("Deleting filesystem")
            delete_fs = scard_commands.DeleteFileSystem()
            delete_fs.transmit(self.card_connection)
        except SPSDKTpError:
            logger.warning("Failed to delete filesystem. This is expected on the very first run")
        logger.info("Creating filesystem")
        create_fs = scard_commands.CreateFileSystem(objects_count=15)
        create_fs.transmit(self.card_connection)

        logger.info("Setting ECC params")
        ecc_params_cmd = scard_commands.SetProvisioningItem(
            prov_item=ProvItem.ECC_DOMAIN_PARAM,
            data=(
                b"\x90 \xff\xff\xff\xff\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
                b"\x00\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\xfc\x91 Z\xc65\xd8\xaa:\x93\xe7"
                b"\xb3\xeb\xbdUv\x98\x86\xbce\x1d\x06\xb0\xccS\xb0\xf6;\xce<>'\xd2`K\x92 \xff\xff"
                b"\xff\xff\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xff\xff"
                b"\xff\xff\xff\xff\xff\xff\xff\xff\xff\xff\x93A\x04k\x17\xd1\xf2\xe1,BG\xf8\xbc\xe6"
                b"\xe5c\xa4@\xf2w\x03}\x81-\xeb3\xa0\xf4\xa19E\xd8\x98\xc2\x96O\xe3B\xe2\xfe\x1a\x7f"
                b"\x9b\x8e\xe7\xebJ|\x0f\x9e\x16+\xce3Wk1^\xce\xcb\xb6@h7\xbfQ\xf5\x95 \xff\xff\xff"
                b"\xff\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xff\xbc\xe6\xfa\xad\xa7\x17\x9e"
                b"\x84\xf3\xb9\xca\xc2\xfcc%Q\x97\x02\x00\x01"
            ),
        )
        ecc_params_cmd.transmit(self.card_connection)

    def seal(self) -> None:
        """Seal the provisioning device."""
        if numberify_version(self.descriptor.version) < numberify_version("1.0.7"):
            raise SPSDKTpError(
                "Card sealing operation is available in TP Applet version 1.0.7+"
                f"; provided applet has version {self.descriptor.version}"
            )
        logger.info("Start sealing the Smart Card")
        finalize_fs = scard_commands.FinalizeFileSystem()
        finalize_fs.transmit(self.card_connection)
        logger.info("Sealing Smart Card completed")

    def get_prov_counter(self) -> int:
        """Get actual provisioning counter."""
        counter_cmd = scard_commands.GetProductionCounter()
        counter = counter_cmd.format(counter_cmd.transmit(self.card_connection))
        return counter

    def get_prov_remainder(self) -> int:
        """Get the number of remaining provisioning attempts."""
        if numberify_version(self.descriptor.version) < numberify_version("1.0.9"):
            raise SPSDKTpError(
                "Prov remainder operation is available in TP Applet version 1.0.9+"
                f"; provided applet has version {self.descriptor.version}"
            )
        remainder_cmd = scard_commands.GetProductionRemainder()
        counter = remainder_cmd.format(remainder_cmd.transmit(self.card_connection))
        return counter

    @staticmethod
    def _serialize_private_key(
        prk_file_path: str, search_dirs: Optional[List[str]] = None
    ) -> bytes:
        prk_file = find_file(prk_file_path, search_paths=search_dirs)
        prk = load_private_key(prk_file)
        assert isinstance(prk, ec.EllipticCurvePrivateKeyWithSerialization)
        d_number = prk.private_numbers().private_value
        d_bytes = d_number.to_bytes(length=32, byteorder="big")
        return d_bytes

    @staticmethod
    def _serialize_cert(
        cert_file_path: str, search_dirs: Optional[List[str]] = None, destination: int = 0
    ) -> bytes:
        cert_file = find_file(cert_file_path, search_paths=search_dirs)
        cert_data = load_binary(cert_file)
        return TpDevSmartCard._serialize_cert_data(cert_data, destination)

    @staticmethod
    def _serialize_cert_data(cert_data: bytes, destination: int = 0) -> bytes:
        cert = load_der_x509_certificate(cert_data)

        scn_raw = cert.subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value
        scn = bytes(scn_raw, encoding="ascii") if isinstance(scn_raw, str) else scn_raw

        # 16B UUID, hyphen, 1B certificate ID
        scn_length = 32 + 1 + 2
        scn_offset = cert_data.index(scn) + len(scn) - scn_length

        pub_key = cert.public_key()
        assert isinstance(pub_key, ec.EllipticCurvePublicKey)
        x = pub_key.public_numbers().x.to_bytes(length=32, byteorder="big")
        y = pub_key.public_numbers().y.to_bytes(length=32, byteorder="big")
        x_offset = cert_data.index(x)
        y_offset = cert_data.index(y)

        sig_offset = cert_data.index(cert.signature)
        sig_length = len(cert.signature)

        tbs_offset = cert_data.index(cert.tbs_certificate_bytes)
        tbs_length = len(cert.tbs_certificate_bytes)

        cert_length = len(cert_data)

        return (
            cert_data
            # final certificate could be 4B longer due to changing signature size
            + bytes(4)
            + scn_offset.to_bytes(length=2, byteorder="big")
            + scn_length.to_bytes(length=2, byteorder="big")
            + x_offset.to_bytes(length=2, byteorder="big")
            + y_offset.to_bytes(length=2, byteorder="big")
            + sig_offset.to_bytes(length=2, byteorder="big")
            + sig_length.to_bytes(length=2, byteorder="big")
            + tbs_offset.to_bytes(length=2, byteorder="big")
            + tbs_length.to_bytes(length=2, byteorder="big")
            + cert_length.to_bytes(length=2, byteorder="big")
            + destination.to_bytes(length=4, byteorder="little")
        )

    @staticmethod
    def _create_oem_cert_template(config_data: dict, config_dir: Optional[str] = None) -> bytes:
        logger.info("Creating OEM ID template")
        cert_config: dict = config_data["oem_id_config"]
        sanitize_common_name(cert_config["subject"])
        issuer = generate_name(cert_config["issuer"])
        subject = generate_name(cert_config["subject"])
        private_key = load_private_key(
            find_file(
                config_data["oem_id_prk_path"], search_paths=[config_dir] if config_dir else None
            )
        )
        assert isinstance(private_key, ec.EllipticCurvePrivateKey)
        public_key = generate_ecc_public_key(private_key=private_key)
        cert = generate_certificate(
            subject=subject,
            issuer=issuer,
            subject_public_key=public_key,
            issuer_private_key=private_key,
            duration=cert_config["duration"],
            if_ca=False,
            serial_number=cert_config.get("serial_number"),
        )
        cert_raw = convert_certificate_into_bytes(cert, Encoding.DER)

        database = Database(TP_DATA_FILE)
        max_size = value_to_int(
            database.get_device_value("max_oem_cert_size", config_data["family"])
        )
        if len(cert_raw) > max_size:
            raise SPSDKTpError(
                f"OEM certificate is too big: {len(cert_raw)}B, max size: {max_size}B! "
                "Please adjust information in SUBJECT and/or ISSUER portion of the certificate."
            )

        return cert_raw



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2016-2018 Martin Olejar
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Helper module for more human-friendly interpretation of the target device properties."""


import ctypes
from copy import deepcopy
from typing import List, Optional, Tuple, Type, Union

from spsdk.mboot.exceptions import McuBootError
from spsdk.utils.easy_enum import Enum

from .commands import CommandTag
from .error_codes import StatusCode
from .memories import ExtMemPropTags, MemoryRegion


########################################################################################################################
# McuBoot helper functions
########################################################################################################################
def size_fmt(value: Union[int, float], kibibyte: bool = True) -> str:
    """Convert size value into string format.

    :param value: The raw value
    :param kibibyte: True if 1024 Bytes represent 1kB or False if 1000 Bytes represent 1kB
    :return: Stringified value
    """
    base, suffix = [(1000.0, "B"), (1024.0, "iB")][kibibyte]
    x = "B"
    for x in ["B"] + [prefix + suffix for prefix in list("kMGTP")]:
        if -base < value < base:
            break
        value /= base

    return f"{value} {x}" if x == "B" else f"{value:3.1f} {x}"


########################################################################################################################
# McuBoot helper classes
########################################################################################################################


class Version:
    """McuBoot current and target version type."""

    def __init__(self, *args: Union[str, int], **kwargs: int):
        """Initialize the Version object.

        :raises McuBootError: Argument passed the not str not int
        """
        self.mark = kwargs.get("mark", "K")
        self.major = kwargs.get("major", 0)
        self.minor = kwargs.get("minor", 0)
        self.fixation = kwargs.get("fixation", 0)
        if args:
            if isinstance(args[0], int):
                self.from_int(args[0])
            elif isinstance(args[0], str):
                self.from_str(args[0])
            else:
                raise McuBootError("Value must be 'str' or 'int' type !")

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Version) and vars(obj) == vars(self)

    def __ne__(self, obj: object) -> bool:
        return not self.__eq__(obj)

    def __lt__(self, obj: "Version") -> bool:
        return self.to_int(True) < obj.to_int(True)

    def __le__(self, obj: "Version") -> bool:
        return self.to_int(True) <= obj.to_int(True)

    def __gt__(self, obj: "Version") -> bool:
        return self.to_int(True) > obj.to_int(True)

    def __ge__(self, obj: "Version") -> bool:
        return self.to_int(True) >= obj.to_int(True)

    def __str__(self) -> str:
        return self.to_str()

    def __repr__(self) -> str:
        return f"<Version(mark={self.mark}, major={self.major}, minor={self.minor}, fixation={self.fixation})>"

    def from_int(self, value: int) -> None:
        """Parse version data from raw int value.

        :param value: Raw integer input
        """
        mark = (value >> 24) & 0xFF
        self.mark = chr(mark) if 64 < mark < 91 else None  # type: ignore
        self.major = (value >> 16) & 0xFF
        self.minor = (value >> 8) & 0xFF
        self.fixation = value & 0xFF

    def from_str(self, value: str) -> None:
        """Parse version data from string value.

        :param value: String representation input
        """
        mark_major, minor, fixation = value.split(".")
        if len(mark_major) > 1 and mark_major[0] not in "0123456789":
            self.mark = mark_major[0]
            self.major = int(mark_major[1:])
        else:
            self.major = int(mark_major)
        self.minor = int(minor)
        self.fixation = int(fixation)

    def to_int(self, no_mark: bool = False) -> int:
        """Get version value in raw integer format.

        :param no_mark: If True, return value without mark
        :return: Integer representation
        """
        value = self.major << 16 | self.minor << 8 | self.fixation
        mark = 0 if no_mark or self.mark is None else ord(self.mark) << 24  # type: ignore
        return value | mark

    def to_str(self, no_mark: bool = False) -> str:
        """Get version value in readable string format.

        :param no_mark: If True, return value without mark
        :return: String representation
        """
        value = f"{self.major}.{self.minor}.{self.fixation}"
        mark = "" if no_mark or self.mark is None else self.mark
        return f"{mark}{value}"


########################################################################################################################
# McuBoot Properties
########################################################################################################################

# fmt: off
class PropertyTag(Enum):
    """McuBoot Properties."""

    # LIST_PROPERTIES            = (0x00, 'ListProperties', 'List Properties')
    CURRENT_VERSION            = (0x01, "CurrentVersion", "Current Version")
    AVAILABLE_PERIPHERALS      = (0x02, "AvailablePeripherals", "Available Peripherals")
    FLASH_START_ADDRESS        = (0x03, "FlashStartAddress", "Flash Start Address")
    FLASH_SIZE                 = (0x04, "FlashSize", "Flash Size")
    FLASH_SECTOR_SIZE          = (0x05, "FlashSectorSize", "Flash Sector Size")
    FLASH_BLOCK_COUNT          = (0x06, "FlashBlockCount", "Flash Block Count")
    AVAILABLE_COMMANDS         = (0x07, "AvailableCommands", "Available Commands")
    CRC_CHECK_STATUS           = (0x08, "CrcCheckStatus", "CRC Check Status")
    LAST_ERROR                 = (0x09, "LastError", "Last Error Value")
    VERIFY_WRITES              = (0x0A, "VerifyWrites", "Verify Writes")
    MAX_PACKET_SIZE            = (0x0B, "MaxPacketSize", "Max Packet Size")
    RESERVED_REGIONS           = (0x0C, "ReservedRegions", "Reserved Regions")
    VALIDATE_REGIONS           = (0x0D, "ValidateRegions", "Validate Regions")
    RAM_START_ADDRESS          = (0x0E, "RamStartAddress", "RAM Start Address")
    RAM_SIZE                   = (0x0F, "RamSize", "RAM Size")
    SYSTEM_DEVICE_IDENT        = (0x10, "SystemDeviceIdent", "System Device Identification")
    FLASH_SECURITY_STATE       = (0x11, "FlashSecurityState", "Security State")
    UNIQUE_DEVICE_IDENT        = (0x12, "UniqueDeviceIdent", "Unique Device Identification")
    FLASH_FAC_SUPPORT          = (0x13, "FlashFacSupport", "Flash Fac. Support")
    FLASH_ACCESS_SEGMENT_SIZE  = (0x14, "FlashAccessSegmentSize", "Flash Access Segment Size",)
    FLASH_ACCESS_SEGMENT_COUNT = (0x15, "FlashAccessSegmentCount", "Flash Access Segment Count",)
    FLASH_READ_MARGIN          = (0x16, "FlashReadMargin", "Flash Read Margin")
    QSPI_INIT_STATUS           = (0x17, "QspiInitStatus", "QuadSPI Initialization Status")
    TARGET_VERSION             = (0x18, "TargetVersion", "Target Version")
    EXTERNAL_MEMORY_ATTRIBUTES = (0x19, "ExternalMemoryAttributes", "External Memory Attributes",)
    RELIABLE_UPDATE_STATUS     = (0x1A, "ReliableUpdateStatus", "Reliable Update Status")
    FLASH_PAGE_SIZE            = (0x1B, "FlashPageSize", "Flash Page Size")
    IRQ_NOTIFIER_PIN           = (0x1C, "IrqNotifierPin", "Irq Notifier Pin")
    PFR_KEYSTORE_UPDATE_OPT    = (0x1D, "PfrKeystoreUpdateOpt", "PFR Keystore Update Opt")
    BYTE_WRITE_TIMEOUT_MS      = (0x1E, "ByteWriteTimeoutMs", "Byte Write Timeout in ms")
    FUSE_LOCKED_STATUS         = (0x1F, "FuseLockedStatus", "Fuse Locked Status")


class PropertyTagKw45xx(Enum):
    """McuBoot Properties."""

    VERIFY_ERASE               = (0x0A, "VerifyErase", "Verify Erase")
    BOOT_STATUS_REGISTER       = (0x14, "BootStatusRegister", "Boot Status Register",)
    FIRMWARE_VERSION           = (0x15, "FirmwareVersion", "Firmware Version",)
    FUSE_PROGRAM_VOLTAGE       = (0x16, "FuseProgramVoltage", "Fuse Program Voltage")

class PeripheryTag(Enum):
    """Tags representing peripherals."""

    UART      = (0x01, "UART", "UART Interface")
    I2C_SLAVE = (0x02, "I2C-Slave", "I2C Slave Interface")
    SPI_SLAVE = (0x04, "SPI-Slave", "SPI Slave Interface")
    CAN       = (0x08, "CAN", "CAN Interface")
    USB_HID   = (0x10, "USB-HID", "USB HID-Class Interface")
    USB_CDC   = (0x20, "USB-CDC", "USB CDC-Class Interface")
    USB_DFU   = (0x40, "USB-DFU", "USB DFU-Class Interface")
    LIN       = (0x80, "LIN", "LIN Interface")


class FlashReadMargin(Enum):
    """Scopes for flash read."""

    NORMAL  = (0, "Normal")
    USER    = (1, "User")
    FACTORY = (2, "Factory")


class PfrKeystoreUpdateOpt(Enum):
    """Options for PFR updating."""

    KEY_PROVISIONING = (0, "KeyProvisioning")
    WRITE_MEMORY     = (1, "WriteMemory")
# fmt: on

########################################################################################################################
# McuBoot Properties Values
########################################################################################################################


class PropertyValueBase:
    """Base class for property value."""

    __slots__ = ("tag", "name", "desc")

    def __init__(self, tag: int, name: Optional[str] = None, desc: Optional[str] = None) -> None:
        """Initialize the base of property.

        :param tag: Property tag, see: `PropertyTag`
        :param name: Optional name for the property
        :param desc: Optional description for the property
        """
        self.tag = tag
        self.name = name or PropertyTag.get(tag, "")
        self.desc = desc or PropertyTag.desc(tag)

    def __str__(self) -> str:
        return f"{self.desc} = {self.to_str()}"

    def to_str(self) -> str:
        """Stringified representation of a property.

        Derived classes should implement this function.

        :return: String representation
        :raises NotImplementedError: Derived class has to implement this method
        """
        raise NotImplementedError("Derived class has to implement this method.")


class IntValue(PropertyValueBase):
    """Integer-based value property."""

    __slots__ = (
        "value",
        "_fmt",
    )

    def __init__(self, tag: int, raw_values: List[int], str_format: str = "dec") -> None:
        """Initialize the integer-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        :param str_format: Format to display the value ('dec', 'hex', 'size')
        """
        super().__init__(tag)
        self._fmt = str_format
        self.value = raw_values[0]

    def to_int(self) -> int:
        """Get the raw integer property representation."""
        return self.value

    def to_str(self) -> str:
        """Get stringified property representation."""
        if self._fmt == "size":
            str_value = size_fmt(self.value)
        elif self._fmt == "hex":
            str_value = f"0x{self.value:08X}"
        elif self._fmt == "dec":
            str_value = str(self.value)
        elif self._fmt == "int32":
            str_value = str(ctypes.c_int32(self.value).value)
        else:
            str_value = self._fmt.format(self.value)
        return str_value


class BoolValue(PropertyValueBase):
    """Boolean-based value property."""

    __slots__ = (
        "value",
        "_true_values",
        "_false_values",
        "_true_string",
        "_false_string",
    )

    def __init__(
        self,
        tag: int,
        raw_values: List[int],
        true_values: Tuple[int] = (1,),
        true_string: str = "YES",
        false_values: Tuple[int] = (0,),
        false_string: str = "NO",
    ) -> None:
        """Initialize the Boolean-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        :param true_values: Values representing 'True', defaults to (1,)
        :param true_string: String representing 'True, defaults to 'YES'
        :param false_values: Values representing 'False', defaults to (0,)
        :param false_string: String representing 'False, defaults to 'NO'
        """
        super().__init__(tag)
        self._true_values = true_values
        self._true_string = true_string
        self._false_values = false_values
        self._false_string = false_string
        self.value = raw_values[0]

    def __bool__(self) -> bool:
        return self.value in self._true_values

    def to_int(self) -> int:
        """Get the raw integer portion of the property."""
        return self.value

    def to_str(self) -> str:
        """Get stringified property representation."""
        return self._true_string if self.value in self._true_values else self._false_string


class EnumValue(PropertyValueBase):
    """Enumeration value property."""

    __slots__ = ("value", "enum", "_na_msg")

    def __init__(
        self,
        tag: int,
        raw_values: List[int],
        enum: Type[Enum],
        na_msg: str = "Unknown Item",
    ) -> None:
        """Initialize the enumeration-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        :param enum: Enumeration to pick from
        :param na_msg: Message to display if an item is not found in the enum
        """
        super().__init__(tag)
        self._na_msg = na_msg
        self.enum = enum
        self.value = raw_values[0]

    def to_int(self) -> int:
        """Get the raw integer portion of the property."""
        return self.value

    def to_str(self) -> str:
        """Get stringified property representation."""
        return self.enum.name(self.value, f"{self._na_msg}: {self.value}")


class VersionValue(PropertyValueBase):
    """Version property class."""

    __slots__ = ("value",)

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the Version-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.value = Version(raw_values[0])

    def to_int(self) -> int:
        """Get the raw integer portion of the property."""
        return self.value.to_int()

    def to_str(self) -> str:
        """Get stringified property representation."""
        return self.value.to_str()


class DeviceUidValue(PropertyValueBase):
    """Device UID value property."""

    __slots__ = ("value",)

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the Version-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.value = b"".join(
            [int.to_bytes(val, length=4, byteorder="little") for val in raw_values]
        )

    def to_int(self) -> int:
        """Get the raw integer portion of the property."""
        return int.from_bytes(self.value, byteorder="big")

    def to_str(self) -> str:
        """Get stringified property representation."""
        return " ".join(f"{item:02X}" for item in self.value)


class ReservedRegionsValue(PropertyValueBase):
    """Reserver Regions property."""

    __slots__ = ("regions",)

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the ReserverRegion-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.regions: List[MemoryRegion] = []
        for i in range(0, len(raw_values), 2):
            if raw_values[i + 1] == 0:
                continue
            self.regions.append(MemoryRegion(raw_values[i], raw_values[i + 1]))

    def __str__(self) -> str:
        return f"{self.desc} =\n{self.to_str()}"

    def to_str(self) -> str:
        """Get stringified property representation."""
        return "\n".join([f"    Region {i}: {region}" for i, region in enumerate(self.regions)])


class AvailablePeripheralsValue(PropertyValueBase):
    """Available Peripherals property."""

    __slots__ = ("value",)

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the AvailablePeripherals-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.value = raw_values[0]

    def to_int(self) -> int:
        """Get the raw integer portion of the property."""
        return self.value

    def to_str(self) -> str:
        """Get stringified property representation."""
        return ", ".join([key for key, value, _ in PeripheryTag if value & self.value])  # type: ignore


class AvailableCommandsValue(PropertyValueBase):
    """Available commands property."""

    __slots__ = ("value",)

    @property
    def tags(self) -> List[str]:
        """List of tags representing Available commands."""
        return [
            tag_value  # type: ignore
            for _, tag_value, _ in CommandTag  # type: ignore
            if tag_value > 0 and (1 << tag_value - 1) & self.value  # type: ignore
        ]

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the AvailableCommands-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.value = raw_values[0]

    def __contains__(self, item: int) -> bool:
        return isinstance(item, int) and bool((1 << item - 1) & self.value)

    def to_str(self) -> str:
        """Get stringified property representation."""
        return [
            name  # type: ignore
            for name, value, _ in CommandTag  # type: ignore
            if value > 0 and (1 << value - 1) & self.value  # type: ignore
        ]


class IrqNotifierPinValue(PropertyValueBase):
    """IRQ notifier pin property."""

    __slots__ = ("value",)

    @property
    def pin(self) -> int:
        """Number of the pin used for reporting IRQ."""
        return self.value & 0xFF

    @property
    def port(self) -> int:
        """Number of the port used for reporting IRQ."""
        return (self.value >> 8) & 0xFF

    @property
    def enabled(self) -> bool:
        """Indicates whether IRQ reporting is enabled."""
        return bool(self.value & (1 << 32))

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the IrqNotifierPin-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.value = raw_values[0]

    def __bool__(self) -> bool:
        return self.enabled

    def to_str(self) -> str:
        """Get stringified property representation."""
        return (
            f"IRQ Port[{self.port}], Pin[{self.pin}] is {'enabled' if self.enabled else 'disabled'}"
        )


class ExternalMemoryAttributesValue(PropertyValueBase):
    """Attributes for external memories."""

    __slots__ = (
        "value",
        "mem_id",
        "start_address",
        "total_size",
        "page_size",
        "sector_size",
        "block_size",
    )

    def __init__(self, tag: int, raw_values: List[int], mem_id: int = 0) -> None:
        """Initialize the ExternalMemoryAttributes-based property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        :param mem_id: ID of the external memory
        """
        super().__init__(tag)
        self.mem_id = mem_id
        self.start_address = raw_values[1] if raw_values[0] & ExtMemPropTags.START_ADDRESS else None
        self.total_size = (
            raw_values[2] * 1024 if raw_values[0] & ExtMemPropTags.SIZE_IN_KBYTES else None
        )
        self.page_size = raw_values[3] if raw_values[0] & ExtMemPropTags.PAGE_SIZE else None
        self.sector_size = raw_values[4] if raw_values[0] & ExtMemPropTags.SECTOR_SIZE else None
        self.block_size = raw_values[5] if raw_values[0] & ExtMemPropTags.BLOCK_SIZE else None
        self.value = raw_values[0]

    def to_str(self) -> str:
        """Get stringified property representation."""
        str_values = []
        if self.start_address is not None:
            str_values.append(f"Start Address: 0x{self.start_address:08X}")
        if self.total_size is not None:
            str_values.append(f"Total Size:    {size_fmt(self.total_size)}")
        if self.page_size is not None:
            str_values.append(f"Page Size:     {size_fmt(self.page_size)}")
        if self.sector_size is not None:
            str_values.append(f"Sector Size:   {size_fmt(self.sector_size)}")
        if self.block_size is not None:
            str_values.append(f"Block Size:    {size_fmt(self.block_size)}")
        return ", ".join(str_values)


class FuseLock:
    """Fuse Lock."""

    def __init__(self, index: int, locked: bool) -> None:
        """Initialize object representing information about fuse lock.

        :param index: value of OTP index
        :param locked: status of the lock, true if locked
        """
        self.index = index
        self.locked = locked

    def __str__(self) -> str:
        status = "LOCKED" if self.locked else "UNLOCKED"
        return f"  FUSE{(self.index):03d}: {status}\r\n"


class FuseLockRegister:
    """Fuse Lock Register."""

    def __init__(self, value: int, index: int, start: int = 0) -> None:
        """Initialize object representing the OTP Controller Program Locked Status.

        :param value: value of the register
        :param index: index of the fuse
        :param start: shift to the start of the register

        """
        self.value = value
        self.index = index
        self.msg = ""
        self.bitfields: List[FuseLock] = []

        shift = 0
        for _ in range(start, 32):
            locked = (value >> shift) & 1
            self.bitfields.append(FuseLock(index + shift, bool(locked)))
            shift += 1

    def __str__(self) -> str:
        """Get stringified property representation."""
        if self.bitfields:
            for bitfield in self.bitfields:
                self.msg += str(bitfield)
        return f"\r\n{self.msg}"


class FuseLockedStatus(PropertyValueBase):
    """Class representing FuseLocked registers."""

    __slots__ = ("fuses",)

    def __init__(self, tag: int, raw_values: List[int]) -> None:
        """Initialize the FuseLockedStatus property object.

        :param tag: Property tag, see: `PropertyTag`
        :param raw_values: List of integers representing the property
        """
        super().__init__(tag)
        self.fuses: List[FuseLockRegister] = []
        idx = 0
        for count, val in enumerate(raw_values):
            start = 0
            if count == 0:
                start = 16
            self.fuses.append(FuseLockRegister(val, idx, start))
            idx += 32
            if count == 0:
                idx -= 16

    def to_str(self) -> str:
        """Get stringified property representation."""
        msg = "\r\n"
        for count, register in enumerate(self.fuses):
            msg += f"OTP Controller Program Locked Status {count} Register: {register}"
        return msg

    def get_fuses(self) -> List[FuseLock]:
        """Get list of fuses bitfield objects.

        :return: list of FuseLockBitfield objects
        """
        fuses = []
        for registers in self.fuses:
            fuses.extend(registers.bitfields)
        return fuses


########################################################################################################################
# McuBoot property response parser
########################################################################################################################

PROPERTIES = {
    PropertyTag.CURRENT_VERSION: {"class": VersionValue, "kwargs": {}},
    PropertyTag.AVAILABLE_PERIPHERALS: {
        "class": AvailablePeripheralsValue,
        "kwargs": {},
    },
    PropertyTag.FLASH_START_ADDRESS: {
        "class": IntValue,
        "kwargs": {"str_format": "hex"},
    },
    PropertyTag.FLASH_SIZE: {"class": IntValue, "kwargs": {"str_format": "size"}},
    PropertyTag.FLASH_SECTOR_SIZE: {
        "class": IntValue,
        "kwargs": {"str_format": "size"},
    },
    PropertyTag.FLASH_BLOCK_COUNT: {"class": IntValue, "kwargs": {"str_format": "dec"}},
    PropertyTag.AVAILABLE_COMMANDS: {"class": AvailableCommandsValue, "kwargs": {}},
    PropertyTag.CRC_CHECK_STATUS: {
        "class": EnumValue,
        "kwargs": {"enum": StatusCode, "na_msg": "Unknown CRC Status code"},
    },
    PropertyTag.VERIFY_WRITES: {
        "class": BoolValue,
        "kwargs": {"true_string": "ON", "false_string": "OFF"},
    },
    PropertyTag.LAST_ERROR: {
        "class": EnumValue,
        "kwargs": {"enum": StatusCode, "na_msg": "Unknown Error"},
    },
    PropertyTag.MAX_PACKET_SIZE: {"class": IntValue, "kwargs": {"str_format": "size"}},
    PropertyTag.RESERVED_REGIONS: {"class": ReservedRegionsValue, "kwargs": {}},
    PropertyTag.VALIDATE_REGIONS: {
        "class": BoolValue,
        "kwargs": {"true_string": "ON", "false_string": "OFF"},
    },
    PropertyTag.RAM_START_ADDRESS: {"class": IntValue, "kwargs": {"str_format": "hex"}},
    PropertyTag.RAM_SIZE: {"class": IntValue, "kwargs": {"str_format": "size"}},
    PropertyTag.SYSTEM_DEVICE_IDENT: {
        "class": IntValue,
        "kwargs": {"str_format": "hex"},
    },
    PropertyTag.FLASH_SECURITY_STATE: {
        "class": BoolValue,
        "kwargs": {
            "true_values": (0x00000000, 0x5AA55AA5),
            "true_string": "UNSECURE",
            "false_values": (0x00000001, 0xC33CC33C),
            "false_string": "SECURE",
        },
    },
    PropertyTag.UNIQUE_DEVICE_IDENT: {"class": DeviceUidValue, "kwargs": {}},
    PropertyTag.FLASH_FAC_SUPPORT: {
        "class": BoolValue,
        "kwargs": {"true_string": "ON", "false_string": "OFF"},
    },
    PropertyTag.FLASH_ACCESS_SEGMENT_SIZE: {
        "class": IntValue,
        "kwargs": {"str_format": "size"},
    },
    PropertyTag.FLASH_ACCESS_SEGMENT_COUNT: {
        "class": IntValue,
        "kwargs": {"str_format": "int32"},
    },
    PropertyTag.FLASH_READ_MARGIN: {
        "class": EnumValue,
        "kwargs": {"enum": FlashReadMargin, "na_msg": "Unknown Margin"},
    },
    PropertyTag.QSPI_INIT_STATUS: {
        "class": EnumValue,
        "kwargs": {"enum": StatusCode, "na_msg": "Unknown Error"},
    },
    PropertyTag.TARGET_VERSION: {"class": VersionValue, "kwargs": {}},
    PropertyTag.EXTERNAL_MEMORY_ATTRIBUTES: {
        "class": ExternalMemoryAttributesValue,
        "kwargs": {"mem_id": None},
    },
    PropertyTag.RELIABLE_UPDATE_STATUS: {
        "class": EnumValue,
        "kwargs": {"enum": StatusCode, "na_msg": "Unknown Error"},
    },
    PropertyTag.FLASH_PAGE_SIZE: {"class": IntValue, "kwargs": {"str_format": "size"}},
    PropertyTag.IRQ_NOTIFIER_PIN: {"class": IrqNotifierPinValue, "kwargs": {}},
    PropertyTag.PFR_KEYSTORE_UPDATE_OPT: {
        "class": EnumValue,
        "kwargs": {"enum": PfrKeystoreUpdateOpt, "na_msg": "Unknown"},
    },
    PropertyTag.BYTE_WRITE_TIMEOUT_MS: {
        "class": IntValue,
        "kwargs": {"str_format": "dec"},
    },
    PropertyTag.FUSE_LOCKED_STATUS: {
        "class": FuseLockedStatus,
        "kwargs": {},
    },
}

PROPERTIES_KW45XX = {
    PropertyTagKw45xx.VERIFY_ERASE: {
        "class": BoolValue,
        "kwargs": {"true_string": "ENABLE", "false_string": "DISABLE"},
    },
    PropertyTagKw45xx.BOOT_STATUS_REGISTER: {
        "class": IntValue,
        "kwargs": {"str_format": "int32"},
    },
    PropertyTagKw45xx.FIRMWARE_VERSION: {
        "class": IntValue,
        "kwargs": {"str_format": "int32"},
    },
    PropertyTagKw45xx.FUSE_PROGRAM_VOLTAGE: {
        "class": BoolValue,
        "kwargs": {
            "true_string": "Over Drive Voltage (2.5 V)",
            "false_string": "Normal Voltage (1.8 V)",
        },
    },
}

PROPERTIES_OVERRIDE = {"kw45xx": PROPERTIES_KW45XX}
PROPERTIES_OVERRIDE = {"k32w1xx": PROPERTIES_KW45XX}

PROPERTY_TAG_OVERRIDE = {"kw45xx": PropertyTagKw45xx}
PROPERTY_TAG_OVERRIDE = {"k32w1xx": PropertyTagKw45xx}


def parse_property_value(
    property_tag: int,
    raw_values: List[int],
    ext_mem_id: Optional[int] = None,
    family: Optional[str] = None,
) -> Optional[PropertyValueBase]:
    """Parse the property value received from the device.

    :param property_tag: Tag representing the property
    :param raw_values: Data received from the device
    :param ext_mem_id: ID of the external memory used to read the property, defaults to None
    :param family: supported family
    :return: Object representing the property
    """
    assert isinstance(property_tag, int)
    assert isinstance(raw_values, list)
    properties_dict = deepcopy(PROPERTIES)
    if family:
        properties_dict.update(PROPERTIES_OVERRIDE[family])  # type: ignore
    if property_tag not in properties_dict.keys():
        return None
    cls = properties_dict[property_tag]["class"]  # type: ignore
    kwargs: dict = properties_dict[property_tag]["kwargs"]  # type: ignore
    if "mem_id" in kwargs:
        kwargs["mem_id"] = ext_mem_id  # type: ignore
    obj = cls(property_tag, raw_values, **kwargs)  # type: ignore
    if family:
        obj.name = PROPERTY_TAG_OVERRIDE[family].name(property_tag)
        obj.desc = PROPERTY_TAG_OVERRIDE[family].desc(property_tag)
    return obj  # type: ignore



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for parsing original elf2sb configuration files."""

from typing import List, Optional

from spsdk import SPSDKError
from spsdk.image import MBIMG_SCH_FILE
from spsdk.utils.crypto import CRYPTO_SCH_FILE
from spsdk.utils.crypto.cert_blocks import find_root_certificates, get_main_cert_index
from spsdk.utils.schema_validator import ValidationSchemas, check_config


class RootOfTrustInfo:  # pylint: disable=too-few-public-methods
    """Filters out Root Of Trust information given to elf2sb application."""

    def __init__(self, data: dict, search_paths: Optional[List[str]] = None) -> None:
        """Create object out of data loaded from elf2sb configuration file.

        :param data: Configuration data.
        :param search_paths: List of paths where to search for the file, defaults to None
        :raises SPSDKError: If not valid configuration is detected.
        """
        # Validate input
        sch_cfg = ValidationSchemas.get_schema_file(MBIMG_SCH_FILE)
        sch_crypto_cfg = ValidationSchemas.get_schema_file(CRYPTO_SCH_FILE)
        val_schemas = [sch_cfg[x] for x in ["cert_prv_key", "signing_root_prv_key"]]
        val_schemas.append(sch_crypto_cfg["certificate_root_keys"])
        check_config(data, val_schemas, search_paths=search_paths)

        self.config_data = data
        self.signature_provider_cfg = data.get("signProvider")

        self.private_key = data.get("mainCertPrivateKeyFile") or data.get(
            "mainRootCertPrivateKeyFile"
        )
        if not self.signature_provider_cfg and not self.private_key:
            raise SPSDKError(
                "One of 'mainCertPrivateKeyFile', 'mainRootCertPrivateKeyFile', 'signProvider' must be specified."
            )
        self.public_keys = find_root_certificates(data)
        data_main_cert_index = get_main_cert_index(data, search_paths=search_paths)
        root_cert_file = data.get(f"rootCertificate{data_main_cert_index}File")
        if not root_cert_file:
            raise SPSDKError(f"rootCertificate{data_main_cert_index}File doesn't exist")
        self.public_key_index = self.public_keys.index(root_cert_file)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module containing various functions/modules used throughout the SPSDK."""

import os

from spsdk import SPSDK_DATA_FOLDER

REGS_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "regs")
UTILS_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "utils")

from .exceptions import (
    SPSDKRegsError,
    SPSDKRegsErrorBitfieldNotFound,
    SPSDKRegsErrorEnumNotFound,
    SPSDKRegsErrorRegisterGroupMishmash,
    SPSDKRegsErrorRegisterNotFound,
)
